#
# This file is part of LitePCIe.
#
# Copyright (c) 2020-2023 Enjoy-Digital <enjoy-digital.fr>
# Copyright (c) 2022 Sylvain Munaut <tnt@246tNt.com>
# SPDX-License-Identifier: BSD-2-Clause

import os

from migen import *

from litex.soc.interconnect.csr import *

from litepcie.common import *
from litepcie.phy.common import *

# USPPCIEPHY ----------------------------------------------------------------------------------------

class USPPCIEPHY(LiteXModule):
    endianness    = "little"
    qword_aligned = False
    def __init__(self, platform, pads, speed="gen3", data_width=64, bar0_size=1*MB, cd="sys", pcie_data_width=None):
        # Streams ----------------------------------------------------------------------------------
        self.req_sink   = stream.Endpoint(phy_layout(data_width))
        self.cmp_sink   = stream.Endpoint(phy_layout(data_width))
        self.req_source = stream.Endpoint(phy_layout(data_width))
        self.cmp_source = stream.Endpoint(phy_layout(data_width))
        self.msi        = stream.Endpoint(msi_layout())

        # Registers --------------------------------------------------------------------------------
        self._link_status = CSRStatus(fields=[
            CSRField("status", size=1, values=[
                ("``0b0``", "Link Down."),
                ("``0b1``", "Link Up."),
            ]),
            CSRField("phy_down", size=1, values=[
                ("``0b0``", "PHY Link Up."),
                ("``0b1``", "PHY Link Down."),
            ]),
            CSRField("phy_status", size=2, values=[
                ("``00b``", "No receivers detected."),
                ("``01b``", "Link training in progress."),
                ("``10b``", "Link up, DL initialization in progress."),
                ("``11b``", "Link up, DL initialization completed."),
            ]),
            CSRField("rate", size=2, values=[
                ("``0b00``", "2.5 GT/s."),
                ("``0b01``", "5.0 GT/s."),
                ("``0b10``", "8.0 GT/s."),
            ]),
            CSRField("width", size=3, values=[
                ("``0b000``", "1-Lane link."),
                ("``0b001``", "2-Lane link."),
                ("``0b010``", "4-Lane link."),
                ("``0b011``", "8-Lane link."),
                ("``0b100``", "16-Lane link."),
            ]),
            CSRField("ltssm", size=6, description="LTSSM State"),
        ])
        self._msi_enable        = CSRStatus(description="MSI Enable Status. ``1``: MSI is enabled.")
        self._bus_master_enable = CSRStatus(description="Bus Mastering Status. ``1``: Bus Mastering enabled.")
        self._max_request_size  = CSRStatus(16, description="Negiotiated Max Request Size (in bytes).")
        self._max_payload_size  = CSRStatus(16, description="Negiotiated Max Payload Size (in bytes).")

        # Parameters/Locals ------------------------------------------------------------------------
        if pcie_data_width is None: pcie_data_width = data_width
        self.platform         = platform
        self.data_width       = data_width
        self.pcie_data_width  = pcie_data_width

        self.id               = Signal(16)
        self.bar0_size        = bar0_size
        self.bar0_mask        = get_bar_mask(bar0_size)
        self.max_request_size = Signal(16)
        self.max_payload_size = Signal(16)

        self.external_hard_ip = False

        # # #

        self.speed  = speed
        self.nlanes = nlanes = len(pads.tx_p)

        assert speed           in ["gen3", "gen4"]
        assert nlanes          in [1, 2, 4, 8, 16]
        assert data_width      in [64, 128, 256, 512]
        assert pcie_data_width in [64, 128, 256, 512]

        # Clocking / Reset -------------------------------------------------------------------------
        self.pcie_refclk    = pcie_refclk    = Signal()
        self.pcie_refclk_gt = pcie_refclk_gt = Signal()
        self.pcie_rst_n     = pcie_rst_n     = Signal(reset=1)
        if hasattr(pads, "rst_n"):
            self.comb += pcie_rst_n.eq(pads.rst_n)
        self.specials += Instance("IBUFDS_GTE4",
            p_REFCLK_HROW_CK_SEL = 0,
            i_CEB   = 0,
            i_I     = pads.clk_p,
            i_IB    = pads.clk_n,
            o_O     = pcie_refclk_gt,
            o_ODIV2 = pcie_refclk
        )
        platform.add_period_constraint(pads.clk_p, 1e9/100e6)
        self.cd_pcie = ClockDomain()

        # TX (FPGA --> HOST) CDC / Data Width Conversion -------------------------------------------
        self.cc_datapath = PHYTXDatapath(
            core_data_width = data_width,
            pcie_data_width = pcie_data_width,
            clock_domain    = cd)
        self.comb += self.cmp_sink.connect(self.cc_datapath.sink)
        s_axis_cc = self.cc_datapath.source

        self.rq_datapath = PHYTXDatapath(
            core_data_width = data_width,
            pcie_data_width = pcie_data_width,
            clock_domain    = cd)
        self.comb += self.req_sink.connect(self.rq_datapath.sink)
        s_axis_rq = self.rq_datapath.source

        # RX (HOST --> FPGA) CDC / Data Width Conversion -------------------------------------------
        self.cq_datapath = PHYRXDatapath(
            core_data_width = data_width,
            pcie_data_width = pcie_data_width,
            clock_domain    = cd)
        m_axis_cq = self.cq_datapath.sink
        self.comb += self.cq_datapath.source.connect(self.req_source)

        self.rc_datapath = PHYRXDatapath(
            core_data_width = data_width,
            pcie_data_width = pcie_data_width,
            clock_domain    = cd)
        m_axis_rc = self.rc_datapath.sink
        self.comb += self.rc_datapath.source.connect(self.cmp_source)

        # MSI CDC (FPGA --> HOST) ------------------------------------------------------------------
        if cd == "pcie":
            cfg_msi = self.msi
        else:
            self.msi_cdc = msi_cdc = stream.ClockDomainCrossing(
                layout          = msi_layout(),
                cd_from         = cd,
                cd_to           = "pcie",
                with_common_rst = True,
            )
            self.comb += self.msi.connect(msi_cdc.sink)
            cfg_msi = msi_cdc.source

        # Hard IP Configuration --------------------------------------------------------------------
        def convert_size(command, size, max_size):
            cases = {}
            value = 128
            for i in range(6):
                cases[i] = size.eq(value)
                value = min(value*2, max_size)
            return Case(command, cases)

        link_status     = Signal()
        link_phy_down   = Signal()
        link_phy_status = Signal(2)
        link_rate       = Signal(2)
        link_width      = Signal(3)
        link_ltssm      = Signal(6)

        msi_enable      = Signal()
        serial_number   = Signal(64)
        bus_number      = Signal(8)
        device_number   = Signal(5)
        function_number = Signal(3)

        cfg_function_status  = Signal(16)
        cfg_max_payload_size = Signal(3)
        cfg_max_read_req     = Signal(3)

        self.sync.pcie += [
            convert_size(cfg_max_read_req,     self.max_request_size, max_size=512),
            convert_size(cfg_max_payload_size, self.max_payload_size, max_size=512),
            self.id.eq(Cat(function_number, device_number, bus_number))
        ]
        self.specials += [
            MultiReg(link_status,     self._link_status.fields.status),
            MultiReg(link_phy_down,   self._link_status.fields.phy_down),
            MultiReg(link_phy_status, self._link_status.fields.phy_status),
            MultiReg(link_rate,       self._link_status.fields.rate),
            MultiReg(link_width,      self._link_status.fields.width),
            MultiReg(link_ltssm,      self._link_status.fields.ltssm),
            MultiReg(cfg_function_status,   self._bus_master_enable.status),
            MultiReg(msi_enable,            self._msi_enable.status),
            MultiReg(self.max_request_size, self._max_request_size.status),
            MultiReg(self.max_payload_size, self._max_payload_size.status)
        ]

        self.m_axis_cq = m_axis_cq
        self.s_axis_cc = s_axis_cc
        self.s_axis_rq = s_axis_rq
        self.m_axis_rc = m_axis_rc

        # Hard IP ----------------------------------------------------------------------------------
        class Open(Signal): pass
        m_axis_rc_tuser = Signal(22)
        m_axis_cq_tuser = Signal(22)
        m_axis_rc_tlast = Signal()
        m_axis_cq_tlast = Signal()
        self.pcie_phy_params = dict(
            # Parameters ---------------------------------------------------------------------------
            p_LINK_CAP_MAX_LINK_WIDTH = nlanes,
            p_C_DATA_WIDTH            = pcie_data_width,
            p_KEEP_WIDTH              = pcie_data_width//8,
            p_PCIE_GT_DEVICE          = "GTY",
            p_PCIE_USE_MODE           = "2.0",

            # PCI Express Interface ----------------------------------------------------------------
            # Clk / Rst
            i_sys_clk     = pcie_refclk,
            i_sys_clk_gt  = pcie_refclk_gt,
            i_sys_rst_n   = pcie_rst_n,

            # TX
            o_pci_exp_txp = pads.tx_p,
            o_pci_exp_txn = pads.tx_n,
            # RX
            i_pci_exp_rxp = pads.rx_p,
            i_pci_exp_rxn = pads.rx_n,

            # AXI-S Interface ----------------------------------------------------------------------
            # Common
            o_user_clk_out          = ClockSignal("pcie"),
            o_user_reset_out        = ResetSignal("pcie"),
            o_user_lnk_up           = link_status,
            o_user_app_rdy          = Open(),

            # (FPGA -> Host) Requester Request
            o_pcie_tfc_nph_av       = Open(2),
            o_pcie_tfc_npd_av       = Open(2),
            o_pcie_rq_tag_av        = Open(2),
            o_pcie_rq_seq_num       = Open(4),
            o_pcie_rq_seq_num_vld   = Open(),
            o_pcie_rq_tag           = Open(6),
            o_pcie_rq_tag_vld       = Open(),
            i_s_axis_rq_tvalid      = s_axis_rq.valid,
            i_s_axis_rq_tlast       = s_axis_rq.last,
            o_s_axis_rq_tready      = s_axis_rq.ready,
            i_s_axis_rq_tdata       = s_axis_rq.dat,
            i_s_axis_rq_tkeep       = s_axis_rq.be,
            i_s_axis_rq_tuser       = Constant(0b0000), # Discontinue, Streaming-AXIS, EP(Poisioning), TP(TLP-Digest)

            # (Host -> FPGA) Completer Request
            i_pcie_cq_np_req        = 1,
            o_pcie_cq_np_req_count  = Open(6),
            o_m_axis_cq_tvalid      = m_axis_cq.valid,
            o_m_axis_cq_tlast       = m_axis_cq_tlast,
            i_m_axis_cq_tready      = m_axis_cq.ready,
            o_m_axis_cq_tdata       = m_axis_cq.dat,
            o_m_axis_cq_tkeep       = m_axis_cq.be,
            o_m_axis_cq_tuser       = m_axis_cq_tuser,

            # (Host -> FPGA) Requester Completion
            o_m_axis_rc_tvalid      = m_axis_rc.valid,
            o_m_axis_rc_tlast       = m_axis_rc_tlast,
            i_m_axis_rc_tready      = m_axis_rc.ready,
            o_m_axis_rc_tdata       = m_axis_rc.dat,
            o_m_axis_rc_tkeep       = m_axis_rc.be,
            o_m_axis_rc_tuser       = m_axis_rc_tuser,

            # (FPGA -> Host) Completer Completion
            i_s_axis_cc_tvalid      = s_axis_cc.valid,
            i_s_axis_cc_tlast       = s_axis_cc.last,
            o_s_axis_cc_tready      = s_axis_cc.ready,
            i_s_axis_cc_tdata       = s_axis_cc.dat,
            i_s_axis_cc_tkeep       = s_axis_cc.be,
            i_s_axis_cc_tuser       = Constant(0b0000), # Discontinue, Streaming-AXIS, EP(Poisioning), TP(TLP-Digest)

            # Management Interface -----------------------------------------------------------------
            o_cfg_mgmt_do           = Open(32),
            o_cfg_mgmt_rd_wr_done   = Open(),
            i_cfg_mgmt_di           = 0,
            i_cfg_mgmt_byte_en      = 0,
            i_cfg_mgmt_dwaddr       = 0,
            i_cfg_mgmt_wr_en        = 0,
            i_cfg_mgmt_rd_en        = 0,

            # Flow Control & Status ----------------------------------------------------------------
            o_cfg_fc_cpld           = Open(12),
            o_cfg_fc_cplh           = Open(8),
            o_cfg_fc_npd            = Open(12),
            o_cfg_fc_nph            = Open(8),
            o_cfg_fc_pd             = Open(12),
            o_cfg_fc_ph             = Open(8),
            i_cfg_fc_sel            = 0, # Use PF0

            # Configuration Tx/Rx Message ----------------------------------------------------------
            o_cfg_msg_received      = Open(),
            o_cfg_msg_received_data = Open(8),
            o_cfg_msg_received_type = Open(5),

            i_cfg_msg_transmit      = 0,
            i_cfg_msg_transmit_data = 0,
            i_cfg_msg_transmit_type = 0,
            o_cfg_msg_transmit_done = Open(),

            # Configuration Control Interface ------------------------------------------------------
            # Hot config
            o_pl_received_hot_rst              = Open(),
            i_pl_transmit_hot_rst              = 0,

            # Indentication & Routing
            i_cfg_dsn                          = serial_number,
            i_cfg_ds_bus_number                = bus_number,
            i_cfg_ds_device_number             = device_number,
            i_cfg_ds_function_number           = function_number,
            i_cfg_ds_port_number               = 0,
            i_cfg_subsys_vend_id               = 0x10ee,

            #  power-down request TLP
            i_cfg_power_state_change_ack       = 0,
            o_cfg_power_state_change_interrupt = Open(),

            # Interrupt Signals (Legacy & MSI) -----------------------------------------------------

            i_cfg_interrupt_int             = 0,
            i_cfg_interrupt_pending         = 0,
            o_cfg_interrupt_sent            = Open(),

            o_cfg_interrupt_msi_enable      = msi_enable,
            i_cfg_interrupt_msi_int_valid   = cfg_msi.valid,
            i_cfg_interrupt_msi_int         = cfg_msi.dat,
            o_cfg_interrupt_msi_sent        = cfg_msi.ready,
            o_cfg_interrupt_msi_fail        = Open(),

            o_cfg_interrupt_msi_mmenable    = Open(12),
            o_cfg_interrupt_msi_mask_update = Open(),
            o_cfg_interrupt_msi_data        = Open(32),
            o_cfg_interrupt_msi_vf_enable   = Open(8),

            # Error Reporting Interface ------------------------------------------------------------
            o_cfg_phy_link_down           = link_phy_down,
            o_cfg_phy_link_status         = link_phy_status,
            o_cfg_negotiated_width        = link_width,
            o_cfg_current_speed           = link_rate,
            o_cfg_max_payload             = cfg_max_payload_size,
            o_cfg_max_read_req            = cfg_max_read_req,
            o_cfg_function_status         = cfg_function_status,
            o_cfg_function_power_state    = Open(12),
            o_cfg_vf_status               = Open(16),
            o_cfg_vf_power_state          = Open(24),
            o_cfg_link_power_state        = Open(2),

            o_cfg_err_cor_out             = Open(),
            o_cfg_err_nonfatal_out        = Open(),
            o_cfg_err_fatal_out           = Open(),
            o_cfg_ltr_enable              = Open(),
            o_cfg_ltssm_state             = link_ltssm,
            o_cfg_rcb_status              = Open(4),
            o_cfg_dpa_substate_change     = Open(4),
            o_cfg_obff_enable             = Open(2),
            o_cfg_pl_status_change        = Open(),

            o_cfg_tph_requester_enable    = Open(4),
            o_cfg_tph_st_mode             = Open(12),
            o_cfg_vf_tph_requester_enable = Open(8),
            o_cfg_vf_tph_st_mode          = Open(24),
        )
        self.comb += [
            m_axis_cq.first.eq(m_axis_cq_tuser[14]),
            m_axis_cq.last.eq (m_axis_cq_tlast),
            m_axis_rc.first.eq(m_axis_rc_tuser[14]),
            m_axis_rc.last.eq (m_axis_rc_tlast),
        ]

    # LTSSM Tracer ---------------------------------------------------------------------------------
    def add_ltssm_tracer(self):
        self.ltssm_tracer = LTSSMTracer(self._link_status.fields.ltssm)

    # Hard IP sources ------------------------------------------------------------------------------
    def add_sources(self, platform, phy_path, phy_filename):
        platform.add_ip(os.path.join(phy_path, phy_filename))
        platform.add_source(os.path.join(phy_path, "pcie_usp_support.v"))

    # External Hard IP -----------------------------------------------------------------------------
    def use_external_hard_ip(self, hard_ip_path, hard_ip_filename):
        self.external_hard_ip = True
        self.add_sources(self.platform, hard_ip_path, hard_ip_filename)

    # Finalize -------------------------------------------------------------------------------------
    def do_finalize(self):
        if not self.external_hard_ip:
            phy_path = "xilinx_usp{}_{}_x{}".format(
                 "_hbm" if isinstance(self, USPHBMPCIEPHY) else "",
                 self.speed,
                 self.nlanes
            )
            self.add_sources(self.platform,
                phy_path     = os.path.join(os.path.abspath(os.path.dirname(__file__)), phy_path),
                phy_filename = "pcie_usp.xci"
            )
        self.specials += Instance("pcie_support", **self.pcie_phy_params)

# USPHBMPCIEPHY ------------------------------------------------------------------------------------

class USPHBMPCIEPHY(USPPCIEPHY): pass



#
# This file is part of LitePCIe.
#
# Copyright (c) 2015-2020 Florent Kermarrec <florent@enjoy-digital.fr>
# SPDX-License-Identifier: BSD-2-Clause

from migen import *

from litex.soc.interconnect import wishbone

from litepcie.common import *

# Helpers ------------------------------------------------------------------------------------------

def map_wishbone_dat(address, data, wishbone_dat, qword_aligned=False):
    return [
        If(qword_aligned,
            If(address[2],
                wishbone_dat.eq(data[:32])
            ).Else(
                wishbone_dat.eq(data[32:])
            )
        ).Else(
            wishbone_dat.eq(data[:32])
        )
    ]

# LitePCIeWishboneMaster ---------------------------------------------------------------------------

class LitePCIeWishboneMaster(Module):
    def __init__(self, endpoint,
        address_decoder = lambda a: 1,
        base_address    = 0x00000000,
        qword_aligned   = False):
        self.wishbone = wishbone.Interface()

        # # #

        port = endpoint.crossbar.get_slave_port(address_decoder)

        self.submodules.fsm = fsm = FSM(reset_state="IDLE")
        fsm.act("IDLE",
            If(port.sink.valid & port.sink.first,
                If(port.sink.we,
                    NextState("DO-WRITE")
                ).Else(
                    NextState("DO-READ")
                )
            ).Else(
                port.sink.ready.eq(1)
            )
        )
        self.sync += [
            self.wishbone.sel.eq(0xf),
            self.wishbone.adr.eq(port.sink.adr[2:] + (base_address >> 2)),
            map_wishbone_dat(
                address       = port.sink.adr,
                data          = port.sink.dat,
                wishbone_dat  = self.wishbone.dat_w,
                qword_aligned = qword_aligned,
            ),
        ]
        fsm.act("DO-WRITE",
            self.wishbone.stb.eq(1),
            self.wishbone.we.eq(1),
            self.wishbone.cyc.eq(1),
            If(self.wishbone.ack,
                port.sink.ready.eq(1),
                NextState("IDLE")
            )
        )
        update_dat = Signal()
        fsm.act("DO-READ",
            self.wishbone.stb.eq(1),
            self.wishbone.we.eq(0),
            self.wishbone.cyc.eq(1),
            If(self.wishbone.ack,
                update_dat.eq(1),
                NextState("ISSUE-READ-COMPLETION")
            )
        )
        self.sync += [
            port.source.first.eq(1),
            port.source.last.eq(1),
            port.source.len.eq(1),
            port.source.err.eq(0),
            port.source.tag.eq(port.sink.tag),
            port.source.adr.eq(port.sink.adr),
            port.source.cmp_id.eq(endpoint.phy.id),
            port.source.req_id.eq(port.sink.req_id),
            If(update_dat,
                port.source.dat.eq(self.wishbone.dat_r)
            )
        ]
        fsm.act("ISSUE-READ-COMPLETION",
            port.source.valid.eq(1),
            If(port.source.ready,
                port.sink.ready.eq(1),
                NextState("IDLE")
            )
        )

class LitePCIeWishboneBridge(LitePCIeWishboneMaster): pass # initial name

# LitePCIeWishboneSlave ----------------------------------------------------------------------------

class LitePCIeWishboneSlave(Module):
    def __init__(self, endpoint, qword_aligned=False):
        self.wishbone = wishbone.Interface()

        # # #

        port = endpoint.crossbar.get_master_port()

        self.submodules.fsm = fsm = FSM(reset_state="IDLE")
        fsm.act("IDLE",
            If(self.wishbone.stb & self.wishbone.cyc,
                If(self.wishbone.we,
                    NextState("ISSUE-WRITE")
                ).Else(
                    NextState("ISSUE-READ")
                )
            )
        )
        self.comb += [
            port.source.channel.eq(port.channel),
            port.source.first.eq(1),
            port.source.last.eq(1),
            port.source.adr[2:].eq(self.wishbone.adr),
            port.source.req_id.eq(endpoint.phy.id),
            port.source.tag.eq(0),
            port.source.len.eq(1),
            port.source.dat.eq(self.wishbone.dat_w),
        ]
        fsm.act("ISSUE-WRITE",
            port.source.valid.eq(1),
            port.source.we.eq(1),
            If(port.source.ready,
                self.wishbone.ack.eq(1),
                NextState("IDLE")
            )
        )
        fsm.act("ISSUE-READ",
            port.source.valid.eq(1),
            port.source.we.eq(0),
            If(port.source.ready,
                NextState("RECEIVE-READ-COMPLETION")
            )
        )
        fsm.act("RECEIVE-READ-COMPLETION",
            port.sink.ready.eq(1),
            If(port.sink.valid & port.sink.first,
                map_wishbone_dat(
                    address       = port.sink.adr,
                    data          = port.sink.dat,
                    wishbone_dat  = self.wishbone.dat_r,
                    qword_aligned = qword_aligned,
                ),
                self.wishbone.ack.eq(1),
                NextState("IDLE")
            )
        )



#
# This file is part of LitePCIe.
#
# Copyright (c) 2015-2018 Florent Kermarrec <florent@enjoy-digital.fr>
# SPDX-License-Identifier: BSD-2-Clause

from migen import *

from litex.gen import *

from litex.soc.interconnect import stream
from litex.soc.interconnect.stream import *
from litex.soc.interconnect.packet import *

# Constants/Helpers --------------------------------------------------------------------------------

KB = 1024
MB = 1024*KB
GB = 1024*MB

def get_bar_mask(size):
    mask = 0
    found = 0
    for i in range(32):
        if size%2:
            found = 1
        if found:
            mask |= (1 << i)
        size = size >> 1
    return mask

# Layouts ------------------------------------------------------------------------------------------

def phy_layout(data_width):
    layout = [
        ("dat", data_width),
        ("be",  data_width//8)
    ]
    return EndpointDescription(layout)

def request_layout(data_width, address_width=32):
    layout = [
        # Request Parameters.
        ("req_id",          16), # Requester ID.
        ("we",               1), # Request type; 0 : Read / 1 : Write.
        ("adr",  address_width), # Request address (In Bytes).
        ("len",             10), # Request length (In Dwords).
        ("tag",              8), # Request tag.

        # Data Stream.
        ("dat", data_width),

        # Internal LitePCIe Routing/Identification.
        ("channel", 8), # Crossbar's channel (Used for internal routing).
        ("user_id", 8), # Packet identification (Used for packet delimitation).
    ]
    return EndpointDescription(layout)

def completion_layout(data_width, address_width=32):
    layout = [
        # Completion Parameters.
        ("req_id",          16), # Requester ID.
        ("cmp_id",          16), # Completion ID.
        ("adr",  address_width), # Completion address (In Bytes).
        ("len",             10), # Completion length (In Dwords).
        ("end",              1), # Completion end (Current packet is the last).
        ("err",              1), # Completion error.
        ("tag",              8), # Completion tag.

        # Data Stream.
        ("dat",     data_width),

        # Internal LitePCIe Routing/Identification.
        ("channel", 8), # Crossbar's channel (Used for internal routing).
        ("user_id", 8)  # Packet identification (Used for packet delimitation).
    ]
    return EndpointDescription(layout)

def msi_layout():
    return [("dat", 8)]

def dma_layout(data_width):
    layout = [("data", data_width)]
    return EndpointDescription(layout)



#
# This file is part of LitePCIe.
#
# Copyright (c) 2015-2022 Florent Kermarrec <florent@enjoy-digital.fr>
# SPDX-License-Identifier: BSD-2-Clause

from migen import *

from litepcie.tlp.common import *

# LitePCIeTLPHeaderExtracter64b --------------------------------------------------------------------

class LitePCIeTLPHeaderExtracter64b(Module):
    def __init__(self):
        self.sink   = sink   = stream.Endpoint(phy_layout(64))
        self.source = source = stream.Endpoint(tlp_raw_layout(64))

        # # #

        first = Signal()
        last  = Signal()
        count = Signal()
        dat   = Signal(64,    reset_less=True)
        be    = Signal(64//8, reset_less=True)
        self.sync += \
            If(sink.valid & sink.ready,
                dat.eq(sink.dat),
                be.eq(sink.be)
            )

        self.submodules.fsm = fsm = FSM(reset_state="IDLE")
        fsm.act("IDLE",
            NextValue(first, 1),
            NextValue(last,  0),
            NextValue(count, 0),
            If(sink.valid, NextState("HEADER"))
        )
        fsm.act("HEADER",
            sink.ready.eq(1),
            If(sink.valid,
                NextValue(count, count + 1),
                NextValue(source.header[32*0:32*1], source.header[32*2:32*3]),
                NextValue(source.header[32*1:32*2], source.header[32*3:32*4]),
                NextValue(source.header[32*2:32*3],      sink.dat[32*0:32*1]),
                NextValue(source.header[32*3:32*4],      sink.dat[32*1:32*2]),
                If(count,
                    If(sink.last, NextValue(last, 1)),
                    NextState("COPY")
                )
            )
        )
        fsm.act("COPY",
            source.valid.eq(sink.valid | last),
            source.first.eq(first),
            source.last.eq(sink.last | last),
            If(source.valid & source.ready,
                NextValue(first, 0),
                sink.ready.eq(1 & ~last), # already acked when last is 1
                If(source.last, NextState("IDLE"))
            )
        )
        self.comb += [
            source.dat[32*0:32*1].eq(     dat[32*1:32*2]),
            source.dat[32*1:32*2].eq(sink.dat[32*0:32*1]),
            source.be[  4*0: 4*1].eq(     be[4*1:4*2]),
            source.be[  4*1: 4*2].eq(sink.be[4*0:4*1])
        ]

# LitePCIeTLPHeaderExtracter128b -------------------------------------------------------------------

class LitePCIeTLPHeaderExtracter128b(Module):
    def __init__(self):
        self.sink   = sink   = stream.Endpoint(phy_layout(128))
        self.source = source = stream.Endpoint(tlp_raw_layout(128))

        # # #

        first = Signal()
        last  = Signal()
        dat   = Signal(128,    reset_less=True)
        be    = Signal(128//8, reset_less=True)
        self.sync += \
            If(sink.valid & sink.ready,
                dat.eq(sink.dat),
                be.eq(sink.be)
            )

        self.submodules.fsm = fsm = FSM(reset_state="IDLE")
        fsm.act("IDLE",
            NextValue(first, 1),
            NextValue(last,  0),
            If(sink.valid,
                NextState("HEADER")
            )
        )
        fsm.act("HEADER",
            sink.ready.eq(1),
            If(sink.valid,
                NextValue(source.header[32*0:32*1], sink.dat[32*0:32*1]),
                NextValue(source.header[32*1:32*2], sink.dat[32*1:32*2]),
                NextValue(source.header[32*2:32*3], sink.dat[32*2:32*3]),
                NextValue(source.header[32*3:32*4], sink.dat[32*3:32*4]),
                If(sink.last,
                    NextValue(last, 1)
                ),
                NextState("COPY")
            )
        )
        fsm.act("COPY",
            source.valid.eq(sink.valid | last),
            source.first.eq(first),
            source.last.eq(sink.last | last),
            If(source.valid & source.ready,
                NextValue(first, 0),
                sink.ready.eq(1 & ~last), # already acked when last is 1
                If(source.last,
                    NextState("IDLE")
                )
            )
        )
        self.comb += [
            source.dat[32*0:32*1].eq(     dat[32*3:32*4]),
            source.dat[32*1:32*2].eq(sink.dat[32*0:32*1]),
            source.dat[32*2:32*3].eq(sink.dat[32*1:32*2]),
            source.dat[32*3:32*4].eq(sink.dat[32*2:32*3]),
            source.be[  4*0: 4*1].eq(        be[4*3:4*4]),
            source.be[  4*1: 4*2].eq(   sink.be[4*0:4*1]),
            source.be[  4*2: 4*3].eq(   sink.be[4*1:4*2]),
            source.be[  4*1: 4*2].eq(   sink.be[4*2:4*3]),
        ]

# LitePCIeTLPHeaderExtracter256b -------------------------------------------------------------------

class LitePCIeTLPHeaderExtracter256b(Module):
    def __init__(self):
        self.sink   = sink   = stream.Endpoint(phy_layout(256))
        self.source = source = stream.Endpoint(tlp_raw_layout(256))

        # # #

        first = Signal()
        last  = Signal()
        dat   = Signal(256,    reset_less=True)
        be    = Signal(256//8, reset_less=True)
        self.sync += \
            If(sink.valid & sink.ready,
                dat.eq(sink.dat),
                be.eq(sink.be)
            )

        self.submodules.fsm = fsm = FSM(reset_state="IDLE")
        fsm.act("IDLE",
            NextValue(first, 1),
            NextValue(last,  0),
            If(sink.valid,
                NextState("HEADER")
            )
        )
        fsm.act("HEADER",
            sink.ready.eq(1),
            If(sink.valid,
                NextValue(source.header[32*0:32*1], sink.dat[32*0:32*1]),
                NextValue(source.header[32*1:32*2], sink.dat[32*1:32*2]),
                NextValue(source.header[32*2:32*3], sink.dat[32*2:32*3]),
                NextValue(source.header[32*3:32*4], sink.dat[32*3:32*4]),
                If(sink.last,
                    NextValue(last, 1)
                ),
                NextState("COPY")
            )
        )
        fsm.act("COPY",
            source.valid.eq(sink.valid | last),
            source.first.eq(first),
            source.last.eq(sink.last | last),
            If(source.valid & source.ready,
                NextValue(first, 0),
                sink.ready.eq(1 & ~last), # already acked when last is 1
                If(source.last,
                    NextState("IDLE")
                )
            )
        )
        self.comb += [
            source.dat[32*0:32*1].eq(     dat[32*3:32*4]),
            source.dat[32*1:32*2].eq(     dat[32*4:32*5]),
            source.dat[32*2:32*3].eq(     dat[32*5:32*6]),
            source.dat[32*3:32*4].eq(     dat[32*6:32*7]),
            source.dat[32*4:32*5].eq(     dat[32*7:32*8]),
            source.dat[32*5:32*6].eq(sink.dat[32*0:32*1]),
            source.dat[32*6:32*7].eq(sink.dat[32*1:32*2]),
            source.dat[32*7:32*8].eq(sink.dat[32*2:32*3]),

            source.be[4*0:4*1].eq(     be[4*3:4*4]),
            source.be[4*1:4*2].eq(     be[4*4:4*5]),
            source.be[4*2:4*3].eq(     be[4*5:4*6]),
            source.be[4*3:4*4].eq(     be[4*6:4*7]),
            source.be[4*4:4*5].eq(     be[4*7:4*8]),
            source.be[4*5:4*6].eq(sink.be[4*0:4*1]),
            source.be[4*6:4*7].eq(sink.be[4*1:4*2]),
            source.be[4*7:4*8].eq(sink.be[4*2:4*3])
        ]

# LitePCIeTLPHeaderExtracter512b -------------------------------------------------------------------

class LitePCIeTLPHeaderExtracter512b(Module):
    def __init__(self):
        self.sink   = sink   = stream.Endpoint(phy_layout(512))
        self.source = source = stream.Endpoint(tlp_raw_layout(512))

        # # #

        first = Signal()
        last  = Signal()
        dat   = Signal(512,    reset_less=True)
        be    = Signal(512//8, reset_less=True)
        self.sync += \
            If(sink.valid & sink.ready,
                dat.eq(sink.dat),
                be.eq(sink.be)
            )

        self.submodules.fsm = fsm = FSM(reset_state="IDLE")
        fsm.act("IDLE",
            NextValue(first, 1),
            NextValue(last,  0),
            If(sink.valid,
                NextState("HEADER")
            )
        )
        fsm.act("HEADER",
            sink.ready.eq(1),
            If(sink.valid,
                NextValue(source.header[32*0:32*1], sink.dat[32*0:32*1]),
                NextValue(source.header[32*1:32*2], sink.dat[32*1:32*2]),
                NextValue(source.header[32*2:32*3], sink.dat[32*2:32*3]),
                NextValue(source.header[32*3:32*4], sink.dat[32*3:32*4]),
                If(sink.last,
                    NextValue(last, 1)
                ),
                NextState("COPY")
            )
        )
        fsm.act("COPY",
            source.valid.eq(sink.valid | last),
            source.first.eq(first),
            source.last.eq(sink.last | last),
            If(source.valid & source.ready,
                NextValue(first, 0),
                sink.ready.eq(1 & ~last), # already acked when last is 1
                If(source.last,
                    NextState("IDLE")
                )
            )
        )
        self.comb += [
            source.dat[ 32*0: 32*1].eq(     dat[ 32*3: 32*4]),
            source.dat[ 32*1: 32*2].eq(     dat[ 32*4: 32*5]),
            source.dat[ 32*2: 32*3].eq(     dat[ 32*5: 32*6]),
            source.dat[ 32*3: 32*4].eq(     dat[ 32*6: 32*7]),
            source.dat[ 32*4: 32*5].eq(     dat[ 32*7: 32*8]),
            source.dat[ 32*5: 32*6].eq(     dat[ 32*8: 32*9]),
            source.dat[ 32*6: 32*7].eq(     dat[ 32*9:32*10]),
            source.dat[ 32*7: 32*8].eq(     dat[32*10:32*11]),
            source.dat[ 32*8: 32*9].eq(     dat[32*11:32*12]),
            source.dat[ 32*9:32*10].eq(     dat[32*12:32*13]),
            source.dat[32*10:32*11].eq(     dat[32*13:32*14]),
            source.dat[32*11:32*12].eq(     dat[32*14:32*15]),
            source.dat[32*12:32*13].eq(     dat[32*15:32*16]),
            source.dat[32*13:32*14].eq(sink.dat[ 32*0: 32*1]),
            source.dat[32*14:32*15].eq(sink.dat[ 32*1: 32*2]),
            source.dat[32*15:32*16].eq(sink.dat[ 32*2: 32*3]),


            source.be[ 4*0: 4*1].eq(     be[ 4*3: 4*4]),
            source.be[ 4*1: 4*2].eq(     be[ 4*4: 4*5]),
            source.be[ 4*2: 4*3].eq(     be[ 4*5: 4*6]),
            source.be[ 4*3: 4*4].eq(     be[ 4*6: 4*7]),
            source.be[ 4*4: 4*5].eq(     be[ 4*7: 4*8]),
            source.be[ 4*5: 4*6].eq(     be[ 4*8: 4*9]),
            source.be[ 4*6: 4*7].eq(     be[ 4*9:4*10]),
            source.be[ 4*7: 4*8].eq(     be[4*10:4*11]),
            source.be[ 4*8: 4*9].eq(     be[4*11:4*12]),
            source.be[ 4*9:4*10].eq(     be[4*12:4*13]),
            source.be[4*10:4*11].eq(     be[4*13:4*14]),
            source.be[4*11:4*12].eq(     be[4*14:4*15]),
            source.be[4*12:4*13].eq(     be[4*15:4*16]),
            source.be[4*13:4*14].eq(sink.be[ 4*0: 4*1]),
            source.be[4*14:4*15].eq(sink.be[ 4*1: 4*2]),
            source.be[4*15:4*16].eq(sink.be[ 4*2: 4*3]),
        ]

# LitePCIeTLPDepacketizer --------------------------------------------------------------------------

class LitePCIeTLPDepacketizer(Module):
    def __init__(self, data_width, endianness, address_mask=0):
        self.sink       = stream.Endpoint(phy_layout(data_width))
        self.req_source = stream.Endpoint(request_layout(data_width))
        self.cmp_source = stream.Endpoint(completion_layout(data_width))

        # # #

        # Extract raw header -----------------------------------------------------------------------
        header_extracter_cls = {
             64 : LitePCIeTLPHeaderExtracter64b,
            128 : LitePCIeTLPHeaderExtracter128b,
            256 : LitePCIeTLPHeaderExtracter256b,
            512 : LitePCIeTLPHeaderExtracter512b,
        }
        header_extracter = header_extracter_cls[data_width]()
        self.submodules += header_extracter
        self.comb += self.sink.connect(header_extracter.sink)
        header = header_extracter.source.header

        # Dispatch data according to fmt/type ------------------------------------------------------
        dispatch_source = stream.Endpoint(tlp_common_layout(data_width))
        dispatch_sinks  = [stream.Endpoint(tlp_common_layout(data_width)) for i in range(3)]
        self.comb += dispatch_sinks[0b00].ready.eq(1) # Always ready when unknown.

        self.comb += [
            dispatch_source.valid.eq(header_extracter.source.valid),
            header_extracter.source.ready.eq(dispatch_source.ready),
            dispatch_source.first.eq(header_extracter.source.first),
            dispatch_source.last.eq(header_extracter.source.last),
            tlp_common_header.decode(header, dispatch_source)
        ]
        self.comb += dword_endianness_swap(
            src        = header_extracter.source.dat,
            dst        = dispatch_source.dat,
            data_width = data_width,
            endianness = endianness,
            mode       = "dat",
        )
        self.comb += dword_endianness_swap(
            src        = header_extracter.source.be,
            dst        = dispatch_source.be,
            data_width = data_width,
            endianness = endianness,
            mode       = "be",
        )
        self.submodules.dispatcher = Dispatcher(dispatch_source, dispatch_sinks)

        fmt_type = Cat(dispatch_source.type, dispatch_source.fmt)
        self.comb += [
            self.dispatcher.sel.eq(0b00),
            If((fmt_type == fmt_type_dict["mem_rd32"]) |
               (fmt_type == fmt_type_dict["mem_wr32"]),
                self.dispatcher.sel.eq(0b01),
            ),
            If((fmt_type == fmt_type_dict["cpld"]) |
               (fmt_type == fmt_type_dict["cpl"]),
               self.dispatcher.sel.eq(0b10),
            ),
        ]

        # Decode TLP request and format local request ----------------------------------------------
        self.tlp_req = tlp_req = stream.Endpoint(tlp_request_layout(data_width))
        self.comb += dispatch_sinks[0b01].connect(tlp_req)
        self.comb += tlp_request_header.decode(header, tlp_req)

        req_type   = Cat(tlp_req.type, tlp_req.fmt)
        req_source = self.req_source
        self.comb += [
            req_source.valid.eq(tlp_req.valid),
            req_source.we.eq(tlp_req.valid & (req_type == fmt_type_dict["mem_wr32"])),
            tlp_req.ready.eq(req_source.ready),
            req_source.first.eq(tlp_req.first),
            req_source.last.eq(tlp_req.last),
            req_source.adr.eq(tlp_req.address & (~address_mask)),
            req_source.len.eq(tlp_req.length),
            req_source.req_id.eq(tlp_req.requester_id),
            req_source.tag.eq(tlp_req.tag),
            req_source.dat.eq(tlp_req.dat)
        ]

        # Decode TLP completion and format local completion ----------------------------------------
        self.tlp_cmp = tlp_cmp = stream.Endpoint(tlp_completion_layout(data_width))
        self.comb += dispatch_sinks[0b10].connect(tlp_cmp)
        self.comb += tlp_completion_header.decode(header, tlp_cmp)

        cmp_source = self.cmp_source
        self.comb += [
            cmp_source.valid.eq(tlp_cmp.valid),
            tlp_cmp.ready.eq(cmp_source.ready),
            cmp_source.first.eq(tlp_cmp.first),
            cmp_source.last.eq(tlp_cmp.last),
            cmp_source.len.eq(tlp_cmp.length),
            cmp_source.end.eq(tlp_cmp.length == (tlp_cmp.byte_count[2:])),
            cmp_source.adr.eq(tlp_cmp.lower_address),
            cmp_source.req_id.eq(tlp_cmp.requester_id),
            cmp_source.cmp_id.eq(tlp_cmp.completer_id),
            cmp_source.err.eq(tlp_cmp.status != 0),
            cmp_source.tag.eq(tlp_cmp.tag),
            cmp_source.dat.eq(tlp_cmp.dat)
        ]



#
# This file is part of LitePCIe.
#
# Copyright (c) 2019 Florent Kermarrec <florent@enjoy-digital.fr>
# SPDX-License-Identifier: BSD-2-Clause

import os

from migen import *
from migen.genlib.cdc import MultiReg

from litex.gen import *

from litex.soc.interconnect.csr import *
from litex.soc.interconnect.avalon import *

from litepcie.common import *

# --------------------------------------------------------------------------------------------------

class C5PCIEPHY(LiteXModule):
    endianness    = "little"
    qword_aligned = True
    def __init__(self, platform, pads, data_width=64, bar0_size=1*MB, cd="sys"):
        # Streams ---------------------------------------------------------------------------------
        self.sink   = stream.Endpoint(phy_layout(data_width))
        self.source = stream.Endpoint(phy_layout(data_width))
        self.msi    = stream.Endpoint(msi_layout())

        # Parameters/Locals ------------------------------------------------------------------------
        self.pads             = pads
        self.platform         = platform
        self.data_width       = data_width

        self.id               = Signal(16, reset_less=True)
        self.bar0_size        = bar0_size
        self.bar0_mask        = get_bar_mask(bar0_size)
        self.max_request_size = Signal(16, reset_less=True)
        self.max_payload_size = Signal(16, reset_less=True)

        self.external_hard_ip = False

        # # #

        pcie_clk                       = Signal()
        pcie_rst_n                     = Signal(reset=1)
        pcie_reconfig_clk              = Signal()
        pcie_coreclkout_hip_clk        = Signal()
        pcie_pld_clk_clk               = Signal()
        pcie_pld_clk_1_clk             = Signal()

        pcie_config_tl_tl_cfg_add      = Signal(4)
        pcie_o_config_tl_tl_cfg_ctl    = Signal(32)
        pcie_hip_rst_serdes_pll_locked = Signal()
        pcie_o_power_mngt_pme_to_sr    = Signal()

         # Clocking ---------------------------------------------------------------------------------
        pcie_refclk = Signal()
        self.specials += Instance("ALT_INBUF_DIFF",
            i_i    = pads.clk_p,
            i_ibar = pads.clk_n,
            o_o    = pcie_refclk
        )

        self.cd_pcie = ClockDomain()

        # TX CDC (FPGA --> HOST) -------------------------------------------------------------------
        if cd == "pcie":
            tx_st = self.sink
        else:
            tx_buffer = stream.Buffer(phy_layout(data_width))
            tx_buffer = ClockDomainsRenamer(cd)(tx_buffer)
            tx_cdc    = stream.AsyncFIFO(phy_layout(data_width), 32)
            tx_cdc    = ClockDomainsRenamer({"write": cd, "read": "pcie"})(tx_cdc)
            self.submodules += tx_buffer, tx_cdc
            self.comb += [
                self.sink.connect(tx_buffer.sink),
                tx_buffer.source.connect(tx_cdc.sink)
            ]
            tx_st = tx_cdc.source

        # RX CDC (HOST --> FPGA) -------------------------------------------------------------------
        if cd == "pcie":
            rx_st = self.source
        else:
            rx_cdc    = stream.AsyncFIFO(phy_layout(data_width), 32)
            rx_cdc    = ClockDomainsRenamer({"write": "pcie", "read": cd})(rx_cdc)
            rx_buffer = stream.Buffer(phy_layout(data_width))
            rx_buffer = ClockDomainsRenamer(cd)(rx_buffer)
            self.submodules += rx_buffer, rx_cdc
            self.comb += [
                rx_cdc.source.connect(rx_buffer.sink),
                rx_buffer.source.connect(self.source)
            ]
            rx_st = rx_cdc.sink

        # MSI CDC (FPGA --> HOST) ------------------------------------------------------------------
        if cd == "pcie":
            cfg_msi = self.msi
        else:
            msi_cdc = stream.AsyncFIFO(msi_layout(), 4)
            msi_cdc = ClockDomainsRenamer({"write": cd, "read": "pcie"})(msi_cdc)
            self.submodules += msi_cdc
            self.comb += self.msi.connect(msi_cdc.sink)
            cfg_msi = msi_cdc.source

        # Hard IP Configuration --------------------------------------------------------------------
        def convert_size(command, size):
            cases = {}
            value = 128
            for i in range(6):
                cases[i] = size.eq(value)
                value = value*2
            return Case(command, cases)

        bus_number      = Signal(8)
        device_number   = Signal(5)
        function_number = Signal(3)
        dcommand        = Signal(16)

        self.bus_number      = bus_number
        self.device_number   = device_number
        self.function_number = function_number

        tl_cfg_add_reg_lsb    = Signal()
        tl_cfg_add_reg2_lsb   = Signal()
        cfgctl_addr_change    = Signal()
        cfgctl_addr_change2   = Signal()
        cfgctl_addr_strobe    = Signal()
        captured_cfg_addr_reg = Signal(4)
        captured_cfg_data_reg = Signal(32)

        self.sync.pcie += [
            convert_size(dcommand[12:15], self.max_request_size),
            convert_size(dcommand[5:8], self.max_payload_size),
            self.id.eq(Cat(function_number, device_number, bus_number))
        ]

        # To capture configuration space Register, register LSB bit of tl_cfg_add
        self.sync.pcie += [
            tl_cfg_add_reg_lsb.eq(pcie_config_tl_tl_cfg_add[0]),
            tl_cfg_add_reg2_lsb.eq(tl_cfg_add_reg_lsb)
        ]
        # Detect the address change to generate a strobe to sample the input 32-bit data
        self.sync.pcie += [
            cfgctl_addr_change.eq(tl_cfg_add_reg_lsb != tl_cfg_add_reg2_lsb),
            cfgctl_addr_change2.eq(cfgctl_addr_change),
            cfgctl_addr_strobe.eq(cfgctl_addr_change2)
        ]
        self.sync.pcie += [
            captured_cfg_addr_reg.eq(pcie_config_tl_tl_cfg_add),
            captured_cfg_data_reg.eq(pcie_o_config_tl_tl_cfg_ctl)
        ]

        # Get dcommand
        self.sync.pcie += [
            If((cfgctl_addr_strobe == 1) & (captured_cfg_addr_reg == 0),
                dcommand.eq(captured_cfg_data_reg[0:16])
            )
        ]
        # Get device_number and bus_number
        self.sync.pcie += [
            If((cfgctl_addr_strobe == 1) & (captured_cfg_addr_reg == 15),
                device_number.eq(captured_cfg_data_reg[0:5]),
                bus_number.eq(captured_cfg_data_reg[5:13])
            )
        ]

        # tl_cfg_add[6:4] should represent function number whose information is being presented on
        # tl_cfg_ctl, but only one function is enabled on IP core  in this case function_number is
        # always 0
        self.comb += function_number.eq(0)

        # Native stream <--> AvalonST --------------------------------------------------------------
        tx_n2av = Native2AvalonST(phy_layout(data_width), latency=2)
        tx_n2av = ClockDomainsRenamer("pcie")(tx_n2av)
        self.comb += tx_st.connect(tx_n2av.sink)
        self.submodules += tx_n2av

        rx_av2n = AvalonST2Native(phy_layout(data_width), latency=2)
        rx_av2n = ClockDomainsRenamer("pcie")(rx_av2n)
        self.comb += rx_av2n.source.connect(rx_st)
        self.submodules += rx_av2n

        tx_avst = tx_n2av.source
        rx_avst = rx_av2n.sink

        # Hard IP ----------------------------------------------------------------------------------
        self.pcie_phy_params = dict(
            # Clocks
            i_refclk_clk         = pcie_refclk,
            i_pld_clk_clk        = ClockSignal("pcie"),
            o_coreclkout_hip_clk = ClockSignal("pcie"),

            # Resets
            i_npor_npor      = 1 if not hasattr(pads, "rst_n") else pads.rst_n,
            i_npor_pin_perst = 1 if not hasattr(pads, "rst_n") else pads.rst_n,

            # Hard IP Reconfiguration
            i_reconfig_clk_clk       = pcie_reconfig_clk,
            i_reconfig_reset_reset_n = pcie_rst_n,

            # Power Management
            i_power_mngt_pme_to_cr = pcie_o_power_mngt_pme_to_sr,
            o_power_mngt_pme_to_sr = pcie_o_power_mngt_pme_to_sr,

            # Config (Configuration space)
            o_config_tl_tl_cfg_ctl = pcie_o_config_tl_tl_cfg_ctl,
            o_config_tl_tl_cfg_add = pcie_config_tl_tl_cfg_add,

             # Control
            o_hip_rst_serdes_pll_locked = pcie_hip_rst_serdes_pll_locked,
            i_hip_rst_pld_core_ready    = pcie_hip_rst_serdes_pll_locked,

            # RX Port
            o_rx_st_valid         = rx_avst.valid,
            o_rx_st_startofpacket = rx_avst.first,
            o_rx_st_endofpacket   = rx_avst.last,
            i_rx_st_ready         = rx_avst.ready,
            o_rx_st_data          = rx_avst.dat,

            # TX Port
            i_tx_st_valid         = tx_avst.valid,
            i_tx_st_startofpacket = tx_avst.first,
            i_tx_st_endofpacket   = tx_avst.last,
            o_tx_st_ready         = tx_avst.ready,
            i_tx_st_data          = tx_avst.dat,

            # Serial IF
            i_hip_serial_rx_in0  = pads.rx_p[0],
            i_hip_serial_rx_in1  = pads.rx_p[1],
            i_hip_serial_rx_in2  = pads.rx_p[2],
            i_hip_serial_rx_in3  = pads.rx_p[3],
            o_hip_serial_tx_out0 = pads.tx_p[0],
            o_hip_serial_tx_out1 = pads.tx_p[1],
            o_hip_serial_tx_out2 = pads.tx_p[2],
            o_hip_serial_tx_out3 = pads.tx_p[3],

            # MSI
            i_int_msi_app_msi_num = 0,
            i_int_msi_app_msi_req = cfg_msi.valid,
            i_int_msi_app_msi_tc  = 0,
            o_int_msi_app_msi_ack = cfg_msi.ready,
            i_int_msi_app_int_sts = 0
        )

    # External Hard IP -----------------------------------------------------------------------------
    def use_external_hard_ip(self, hard_ip_path):
        self.external_hard_ip = True
        self.platform.add_source(
            os.path.join("altera", "cyclone_v", "pcie_phy", "synthesis", "pcie_phy.qip"), "QIP")

    def do_finalize(self):
        if not self.external_hard_ip:
            raise ValueError("User needs to provide Hard IP source path with use_external_hard_ip")
        self.specials += Instance("pcie_phy", **self.pcie_phy_params)



#
# This file is part of LitePCIe.
#
# Copyright (c) 2015-2022 Florent Kermarrec <florent@enjoy-digital.fr>
# SPDX-License-Identifier: BSD-2-Clause

from migen import *
from migen.genlib.misc import chooser

from litepcie.tlp.common import *

# LitePCIeTLPHeaderInserter ------------------------------------------------------------------------

class LitePCIeTLPHeaderInserter3DWs4DWs(Module):
    def __init__(self, data_width, header_inserter_3dws_cls, header_inserter_4dws_cls, fmt):
        self.sink   = sink   = stream.Endpoint(tlp_raw_layout(data_width))
        self.source = source = stream.Endpoint(phy_layout(data_width))

        # # #

        # Header Inserters Modules.
        header_inserter_3dws = header_inserter_3dws_cls()
        header_inserter_4dws = header_inserter_4dws_cls()
        self.submodules += header_inserter_3dws, header_inserter_4dws

        # Header Inserters Sel.
        _3DWS_SEL = 0b0
        _4DWS_SEL = 0b1
        header_sel = Signal()
        self.comb += Case(fmt, {
            fmt_dict["mem_rd32"] : header_sel.eq(_3DWS_SEL),
            fmt_dict["mem_rd64"] : header_sel.eq(_4DWS_SEL),
            fmt_dict["mem_wr32"] : header_sel.eq(_3DWS_SEL),
            fmt_dict["mem_wr64"] : header_sel.eq(_4DWS_SEL),
            fmt_dict[    "cpld"] : header_sel.eq(_3DWS_SEL),
            fmt_dict[     "cpl"] : header_sel.eq(_3DWS_SEL),
        })

        # Header Inserters Mux.
        self.comb += Case(header_sel, {
            _3DWS_SEL : [
                sink.connect(header_inserter_3dws.sink),
                header_inserter_3dws.source.connect(source),
            ],
            _4DWS_SEL : [
                sink.connect(header_inserter_4dws.sink),
                header_inserter_4dws.source.connect(source),
            ],
        })

# LitePCIeTLPHeaderInserter64b ---------------------------------------------------------------------

class LitePCIeTLPHeaderInserter64b3DWs(Module):
    def __init__(self):
        self.sink   = sink   = stream.Endpoint(tlp_raw_layout(64))
        self.source = source = stream.Endpoint(phy_layout(64))

        # # #

        count = Signal()
        dat   = Signal(64,    reset_less=True)
        be    = Signal(64//8, reset_less=True)
        last  = Signal(       reset_less=True)
        self.sync += [
            If(sink.valid & sink.ready,
                dat.eq(sink.dat),
                be.eq(sink.be),
                last.eq(sink.last)
            )
        ]

        self.submodules.fsm = fsm = FSM(reset_state="HEADER")
        fsm.act("HEADER",
            sink.ready.eq(1),
            If(sink.valid & sink.first,
                sink.ready.eq(0),
                source.valid.eq(1),
                source.first.eq((count == 0) & sink.first),
                source.last.eq( (count == 1) & sink.last & (sink.be[4*1:] == 0)),
                If(count == 0,
                    source.dat[32*0:32*1].eq(sink.header[32*0:]),
                    source.dat[32*1:32*2].eq(sink.header[32*1:]),
                    source.be[4*0:4*1].eq(0xf),
                    source.be[4*1:4*2].eq(0xf),
                ),
                If(count == 1,
                    source.dat[32*0:32*1].eq(sink.header[32*2:]),
                    source.dat[32*1:32*2].eq(sink.dat[32*0:]),
                    source.be[4*0:4*1].eq(0xf),
                    source.be[4*1:4*2].eq(sink.be[4*0:]),
                ),
                If(source.valid & source.ready,
                    NextValue(count, count + 1),
                    If(count == 1,
                        sink.ready.eq(1),
                        If(~source.last,
                            NextState("DATA")
                        )
                    )
                )
            )
        )
        fsm.act("DATA",
            source.valid.eq(sink.valid | last),
            source.last.eq(last),

            source.dat[32*0:32*1].eq(dat[32*1:]),
            source.dat[32*1:32*2].eq(sink.dat[32*0:]),

            source.be[4*0:4*1].eq(be[4*0:]),
            If(last,
                source.be[4*1:4*2].eq(0x0)
            ).Else(
                source.be[4*1:4*2].eq(sink.be[4*0:])
            ),

            If(source.valid & source.ready,
                sink.ready.eq(~last),
                If(source.last,
                    NextState("HEADER")
                )
            )
        )

class LitePCIeTLPHeaderInserter64b4DWs(Module):
    def __init__(self):
        self.sink   = sink   = stream.Endpoint(tlp_raw_layout(64))
        self.source = source = stream.Endpoint(phy_layout(64))

        # # #

        count = Signal()
        self.submodules.fsm = fsm = FSM(reset_state="HEADER")
        fsm.act("HEADER",
            sink.ready.eq(1),
            If(sink.valid & sink.first,
                sink.ready.eq(0),
                source.valid.eq(1),
                source.first.eq((count == 0) & sink.first),
                source.last.eq( (count == 1) & sink.last & (sink.be == 0)),
                If(count == 0,
                    source.dat[32*0:32*1].eq(sink.header[32*0:]),
                    source.dat[32*1:32*2].eq(sink.header[32*1:]),
                    source.be[4*0:4*1].eq(0xf),
                    source.be[4*1:4*2].eq(0xf),
                ),
                If(count == 1,
                    source.dat[32*0:32*1].eq(sink.header[32*2:]),
                    source.dat[32*1:32*2].eq(sink.header[32*3:]),
                    source.be[4*0:4*1].eq(0xf),
                    source.be[4*1:4*2].eq(0xf),
                ),
                If(source.valid & source.ready,
                    NextValue(count, count + 1),
                    If(count == 1,
                        sink.ready.eq(1),
                        If(~source.last,
                            sink.ready.eq(0),
                            NextState("DATA")
                        )
                    )
                )
            )
        )
        fsm.act("DATA",
            source.valid.eq(sink.valid),
            source.last.eq(sink.last),

            source.dat[32*0:32*1].eq(sink.dat[32*0:]),
            source.dat[32*1:32*2].eq(sink.dat[32*1:]),
            source.be[4*0:4*1].eq(sink.be[4*0:]),
            source.be[4*1:4*2].eq(sink.be[4*1:]),

            If(source.valid & source.ready,
                sink.ready.eq(1),
                If(source.last,
                    NextState("HEADER")
                )
            )
        )

class LitePCIeTLPHeaderInserter64b(LitePCIeTLPHeaderInserter3DWs4DWs):
    def __init__(self, fmt):
        LitePCIeTLPHeaderInserter3DWs4DWs.__init__(self,
            data_width               = 64,
            header_inserter_3dws_cls = LitePCIeTLPHeaderInserter64b3DWs,
            header_inserter_4dws_cls = LitePCIeTLPHeaderInserter64b4DWs,
            fmt                      = fmt,
        )

# LitePCIeTLPHeaderInserter128b --------------------------------------------------------------------

class LitePCIeTLPHeaderInserter128b3DWs(Module):
    def __init__(self):
        self.sink   = sink   = stream.Endpoint(tlp_raw_layout(128))
        self.source = source = stream.Endpoint(phy_layout(128))

        # # #

        dat  = Signal(128,    reset_less=True)
        be   = Signal(128//8, reset_less=True)
        last = Signal(        reset_less=True)
        self.sync += [
            If(sink.valid & sink.ready,
                dat.eq(sink.dat),
                be.eq(sink.be),
                last.eq(sink.last)
            )
        ]

        self.submodules.fsm = fsm = FSM(reset_state="HEADER")
        fsm.act("HEADER",
            sink.ready.eq(1),
            If(sink.valid & sink.first,
                sink.ready.eq(0),
                source.valid.eq(1),
                source.first.eq(1),
                source.last.eq(sink.last & (sink.be[4*1:] == 0)),

                source.dat[32*0:32*1].eq(sink.header[32*0:]),
                source.dat[32*1:32*2].eq(sink.header[32*1:]),
                source.dat[32*2:32*3].eq(sink.header[32*2:]),
                source.dat[32*3:32*4].eq(sink.dat[32*0:]),

                source.be[4*0:4*1].eq(0xf),
                source.be[4*1:4*2].eq(0xf),
                source.be[4*2:4*3].eq(0xf),
                source.be[4*3:4*4].eq(sink.be[0:]),

                If(source.valid & source.ready,
                    sink.ready.eq(1),
                    If(~source.last,
                        NextState("DATA"),
                    )
                )
            )
        )
        fsm.act("DATA",
            source.valid.eq(sink.valid | last),
            source.last.eq(last),

            source.dat[32*0:32*1].eq(dat[32*1:]),
            source.dat[32*1:32*2].eq(dat[32*2:]),
            source.dat[32*2:32*3].eq(dat[32*3:]),
            source.dat[32*3:32*4].eq(sink.dat[32*0:]),

            source.be[4*0:4*1].eq(be[1:]),
            source.be[4*1:4*2].eq(be[2:]),
            source.be[4*2:4*3].eq(be[3:]),
            If(last,
                source.be[4*3:4*4].eq(0x0)
            ).Else(
                source.be[4*3:4*4].eq(sink.be[0:])
            ),

            If(source.valid & source.ready,
                sink.ready.eq(~last),
                If(source.last,
                    NextState("HEADER")
                )
            )
        )

class LitePCIeTLPHeaderInserter128b4DWs(Module):
    def __init__(self):
        self.sink   = sink   = stream.Endpoint(tlp_raw_layout(128))
        self.source = source = stream.Endpoint(phy_layout(128))

        # # #

        self.submodules.fsm = fsm = FSM(reset_state="HEADER")
        fsm.act("HEADER",
            sink.ready.eq(1),
            If(sink.valid & sink.first,
                sink.ready.eq(0),
                source.valid.eq(1),
                source.first.eq(1),
                source.last.eq(sink.last & (sink.be == 0)),

                source.dat[32*0:32*1].eq(sink.header[32*0:]),
                source.dat[32*1:32*2].eq(sink.header[32*1:]),
                source.dat[32*2:32*3].eq(sink.header[32*2:]),
                source.dat[32*3:32*4].eq(sink.header[32*3:]),

                source.be[4*0:4*1].eq(0xf),
                source.be[4*1:4*2].eq(0xf),
                source.be[4*2:4*3].eq(0xf),
                source.be[4*3:4*4].eq(0xf),

                If(source.valid & source.ready,
                    sink.ready.eq(1),
                    If(~source.last,
                        sink.ready.eq(0),
                        NextState("DATA"),
                    )
                )
            )
        )
        fsm.act("DATA",
            source.valid.eq(sink.valid),
            source.last.eq(sink.last),

            source.dat[32*0:32*1].eq(sink.dat[32*0:]),
            source.dat[32*1:32*2].eq(sink.dat[32*1:]),
            source.dat[32*2:32*3].eq(sink.dat[32*2:]),
            source.dat[32*3:32*4].eq(sink.dat[32*3:]),

            source.be[4*0:4*1].eq(sink.be[4*0:]),
            source.be[4*1:4*2].eq(sink.be[4*1:]),
            source.be[4*2:4*3].eq(sink.be[4*2:]),
            source.be[4*3:4*4].eq(sink.be[4*3:]),

            If(source.valid & source.ready,
                sink.ready.eq(1),
                If(source.last,
                    NextState("HEADER")
                )
            )
        )

class LitePCIeTLPHeaderInserter128b(LitePCIeTLPHeaderInserter3DWs4DWs):
    def __init__(self, fmt):
        LitePCIeTLPHeaderInserter3DWs4DWs.__init__(self,
            data_width               = 128,
            header_inserter_3dws_cls = LitePCIeTLPHeaderInserter128b3DWs,
            header_inserter_4dws_cls = LitePCIeTLPHeaderInserter128b4DWs,
            fmt                      = fmt,
        )

# LitePCIeTLPHeaderInserter256b --------------------------------------------------------------------

class LitePCIeTLPHeaderInserter256b3DWs(Module):
    def __init__(self):
        self.sink   = sink   = stream.Endpoint(tlp_raw_layout(256))
        self.source = source = stream.Endpoint(phy_layout(256))

        # # #

        dat  = Signal(256,    reset_less=True)
        be   = Signal(256//8, reset_less=True)
        last = Signal(        reset_less=True)
        self.sync += [
            If(sink.valid & sink.ready,
                dat.eq(sink.dat),
                be.eq(sink.be),
                last.eq(sink.last)
            )
        ]

        self.submodules.fsm = fsm = FSM(reset_state="HEADER")
        fsm.act("HEADER",
            sink.ready.eq(1),
            If(sink.valid & sink.first,
                sink.ready.eq(0),
                source.valid.eq(1),
                source.first.eq(1),
                source.last.eq(sink.last & (sink.be[4*5:] == 0)),

                source.dat[32*0:32*1].eq(sink.header[32*0:]),
                source.dat[32*1:32*2].eq(sink.header[32*1:]),
                source.dat[32*2:32*3].eq(sink.header[32*2:]),
                source.dat[32*3:32*4].eq(sink.dat[32*0:]),
                source.dat[32*4:32*5].eq(sink.dat[32*1:]),
                source.dat[32*5:32*6].eq(sink.dat[32*2:]),
                source.dat[32*6:32*7].eq(sink.dat[32*3:]),
                source.dat[32*7:32*8].eq(sink.dat[32*4:]),

                source.be[4*0:4*1].eq(0xf),
                source.be[4*1:4*2].eq(0xf),
                source.be[4*2:4*3].eq(0xf),
                source.be[4*3:4*4].eq(sink.be[4*0:]),
                source.be[4*4:4*5].eq(sink.be[4*1:]),
                source.be[4*5:4*6].eq(sink.be[4*2:]),
                source.be[4*6:4*7].eq(sink.be[4*3:]),
                source.be[4*7:4*8].eq(sink.be[4*4:]),

                If(source.valid & source.ready,
                    sink.ready.eq(1),
                    If(~source.last,
                        NextState("DATA"),
                    )
                )
            )
        )
        fsm.act("DATA",
            source.valid.eq(sink.valid | last),
            source.last.eq(last),

            source.dat[32*0:32*1].eq(dat[32*5:]),
            source.dat[32*1:32*2].eq(dat[32*6:]),
            source.dat[32*2:32*3].eq(dat[32*7:]),
            source.dat[32*3:32*4].eq(sink.dat[32*0:]),
            source.dat[32*4:32*5].eq(sink.dat[32*1:]),
            source.dat[32*5:32*6].eq(sink.dat[32*2:]),
            source.dat[32*6:32*7].eq(sink.dat[32*3:]),
            source.dat[32*7:32*8].eq(sink.dat[32*4:]),

            source.be[4*0:4*1].eq(be[4*5:]),
            source.be[4*1:4*2].eq(be[4*6:]),
            source.be[4*2:4*3].eq(be[4*7:]),
            If(last,
                source.be[4*3:4*8].eq(0x0)
            ).Else(
                source.be[4*3:4*4].eq(sink.be[4*0:]),
                source.be[4*4:4*5].eq(sink.be[4*1:]),
                source.be[4*5:4*6].eq(sink.be[4*2:]),
                source.be[4*6:4*7].eq(sink.be[4*3:]),
                source.be[4*7:4*8].eq(sink.be[4*4:]),
            ),
            If(source.valid & source.ready,
                sink.ready.eq(~last),
                If(source.last,
                    NextState("HEADER")
                )
            )
        )

class LitePCIeTLPHeaderInserter256b4DWs(Module):
    def __init__(self):
        self.sink   = sink   = stream.Endpoint(tlp_raw_layout(256))
        self.source = source = stream.Endpoint(phy_layout(256))

        # # #

        dat  = Signal(256,    reset_less=True)
        be   = Signal(256//8, reset_less=True)
        last = Signal(        reset_less=True)
        self.sync += [
            If(sink.valid & sink.ready,
                dat.eq(sink.dat),
                be.eq(sink.be),
                last.eq(sink.last)
            )
        ]

        self.submodules.fsm = fsm = FSM(reset_state="HEADER")
        fsm.act("HEADER",
            sink.ready.eq(1),
            If(sink.valid & sink.first,
                sink.ready.eq(0),
                source.valid.eq(1),
                source.first.eq(1),
                source.last.eq(sink.last & (sink.be[4*4:] == 0)),

                source.dat[32*0:32*1].eq(sink.header[32*0:]),
                source.dat[32*1:32*2].eq(sink.header[32*1:]),
                source.dat[32*2:32*3].eq(sink.header[32*2:]),
                source.dat[32*3:32*4].eq(sink.header[32*3:]),
                source.dat[32*4:32*5].eq(sink.dat[32*0:]),
                source.dat[32*5:32*6].eq(sink.dat[32*1:]),
                source.dat[32*6:32*7].eq(sink.dat[32*2:]),
                source.dat[32*7:32*8].eq(sink.dat[32*3:]),

                source.be[4*0:4*1].eq(0xf),
                source.be[4*1:4*2].eq(0xf),
                source.be[4*2:4*3].eq(0xf),
                source.be[4*3:4*4].eq(0xf),
                source.be[4*4:4*5].eq(sink.be[4*0:]),
                source.be[4*5:4*6].eq(sink.be[4*1:]),
                source.be[4*6:4*7].eq(sink.be[4*2:]),
                source.be[4*7:4*8].eq(sink.be[4*3:]),

                If(source.valid & source.ready,
                    sink.ready.eq(1),
                    If(~source.last,
                        NextState("DATA"),
                    )
                )
            )
        )
        fsm.act("DATA",
            source.valid.eq(sink.valid | last),
            source.last.eq(last),

            source.dat[32*0:32*1].eq(dat[32*4:]),
            source.dat[32*1:32*2].eq(dat[32*5:]),
            source.dat[32*2:32*3].eq(dat[32*6:]),
            source.dat[32*3:32*4].eq(dat[32*7:]),
            source.dat[32*4:32*5].eq(sink.dat[32*0:]),
            source.dat[32*5:32*6].eq(sink.dat[32*1:]),
            source.dat[32*6:32*7].eq(sink.dat[32*2:]),
            source.dat[32*7:32*8].eq(sink.dat[32*3:]),

            source.be[4*0:4*1].eq(be[4*4:]),
            source.be[4*1:4*2].eq(be[4*5:]),
            source.be[4*2:4*3].eq(be[4*6:]),
            source.be[4*3:4*4].eq(be[4*7:]),
            If(last,
                source.be[4*4:4*8].eq(0x0)
            ).Else(
                source.be[4*4:4*5].eq(sink.be[4*0:]),
                source.be[4*5:4*6].eq(sink.be[4*1:]),
                source.be[4*6:4*7].eq(sink.be[4*2:]),
                source.be[4*7:4*8].eq(sink.be[4*3:]),
            ),
            If(source.valid & source.ready,
                sink.ready.eq(~last),
                If(source.last,
                    NextState("HEADER")
                )
            )
        )

class LitePCIeTLPHeaderInserter256b(LitePCIeTLPHeaderInserter3DWs4DWs):
    def __init__(self, fmt):
        LitePCIeTLPHeaderInserter3DWs4DWs.__init__(self,
            data_width               = 256,
            header_inserter_3dws_cls = LitePCIeTLPHeaderInserter256b3DWs,
            header_inserter_4dws_cls = LitePCIeTLPHeaderInserter256b4DWs,
            fmt                      = fmt,
        )

# LitePCIeTLPHeaderInserter512b --------------------------------------------------------------------

class LitePCIeTLPHeaderInserter512b3DWs(Module):
    def __init__(self):
        self.sink   = sink   = stream.Endpoint(tlp_raw_layout(512))
        self.source = source = stream.Endpoint(phy_layout(512))

        # # #

        dat  = Signal(512,    reset_less=True)
        be   = Signal(512//8, reset_less=True)
        last = Signal(        reset_less=True)
        self.sync += [
            If(sink.valid & sink.ready,
                dat.eq(sink.dat),
                be.eq(sink.be),
                last.eq(sink.last)
            )
        ]

        self.submodules.fsm = fsm = FSM(reset_state="HEADER")
        fsm.act("HEADER",
            sink.ready.eq(1),
            If(sink.valid & sink.first,
                sink.ready.eq(0),
                source.valid.eq(1),
                source.first.eq(1),
                source.last.eq(sink.last & (sink.be[4*13:] == 0)),

                source.dat[32* 0:32* 1].eq(sink.header[32*0:]),
                source.dat[32* 1:32* 2].eq(sink.header[32*1:]),
                source.dat[32* 2:32* 3].eq(sink.header[32*2:]),
                source.dat[32* 3:32* 4].eq(sink.dat[32* 0:]),
                source.dat[32* 4:32* 5].eq(sink.dat[32* 1:]),
                source.dat[32* 5:32* 6].eq(sink.dat[32* 2:]),
                source.dat[32* 6:32* 7].eq(sink.dat[32* 3:]),
                source.dat[32* 7:32* 8].eq(sink.dat[32* 4:]),
                source.dat[32* 8:32* 9].eq(sink.dat[32* 5:]),
                source.dat[32* 9:32*10].eq(sink.dat[32* 6:]),
                source.dat[32*10:32*11].eq(sink.dat[32* 7:]),
                source.dat[32*11:32*12].eq(sink.dat[32* 8:]),
                source.dat[32*12:32*13].eq(sink.dat[32* 9:]),
                source.dat[32*13:32*14].eq(sink.dat[32*10:]),
                source.dat[32*14:32*15].eq(sink.dat[32*11:]),
                source.dat[32*15:32*16].eq(sink.dat[32*12:]),

                source.be[4* 0:4* 1].eq(0xf),
                source.be[4* 1:4* 2].eq(0xf),
                source.be[4* 2:4* 3].eq(0xf),
                source.be[4* 3:4* 4].eq(sink.be[4* 0:]),
                source.be[4* 4:4* 5].eq(sink.be[4* 1:]),
                source.be[4* 5:4* 6].eq(sink.be[4* 2:]),
                source.be[4* 6:4* 7].eq(sink.be[4* 3:]),
                source.be[4* 7:4* 8].eq(sink.be[4* 4:]),
                source.be[4* 8:4* 9].eq(sink.be[4* 5:]),
                source.be[4* 9:4*10].eq(sink.be[4* 6:]),
                source.be[4*10:4*11].eq(sink.be[4* 7:]),
                source.be[4*11:4*12].eq(sink.be[4* 8:]),
                source.be[4*12:4*13].eq(sink.be[4* 9:]),
                source.be[4*13:4*14].eq(sink.be[4*10:]),
                source.be[4*14:4*15].eq(sink.be[4*11:]),
                source.be[4*15:4*16].eq(sink.be[4*12:]),

                If(source.valid & source.ready,
                    sink.ready.eq(1),
                    If(~source.last,
                        NextState("DATA"),
                    )
                )
            )
        )
        fsm.act("DATA",
            source.valid.eq(sink.valid | last),
            source.last.eq(last),

            source.dat[32* 0:32* 1].eq(     dat[32*13:]),
            source.dat[32* 1:32* 2].eq(     dat[32*14:]),
            source.dat[32* 2:32* 3].eq(     dat[32*15:]),
            source.dat[32* 3:32* 4].eq(sink.dat[32* 0:]),
            source.dat[32* 4:32* 5].eq(sink.dat[32* 1:]),
            source.dat[32* 5:32* 6].eq(sink.dat[32* 2:]),
            source.dat[32* 6:32* 7].eq(sink.dat[32* 3:]),
            source.dat[32* 7:32* 8].eq(sink.dat[32* 4:]),
            source.dat[32* 8:32* 9].eq(sink.dat[32* 5:]),
            source.dat[32* 9:32*10].eq(sink.dat[32* 6:]),
            source.dat[32*10:32*11].eq(sink.dat[32* 7:]),
            source.dat[32*11:32*12].eq(sink.dat[32* 8:]),
            source.dat[32*12:32*13].eq(sink.dat[32* 9:]),
            source.dat[32*13:32*14].eq(sink.dat[32*10:]),
            source.dat[32*14:32*15].eq(sink.dat[32*11:]),
            source.dat[32*15:32*16].eq(sink.dat[32*12:]),

            source.be[4*0:4*1].eq(be[4*13:]),
            source.be[4*1:4*2].eq(be[4*14:]),
            source.be[4*2:4*3].eq(be[4*15:]),
            If(last,
                source.be[4*3:4*16].eq(0x0)
            ).Else(
                source.be[4* 3:4* 4].eq(sink.be[4* 0:]),
                source.be[4* 4:4* 5].eq(sink.be[4* 1:]),
                source.be[4* 5:4* 6].eq(sink.be[4* 2:]),
                source.be[4* 6:4* 7].eq(sink.be[4* 3:]),
                source.be[4* 7:4* 8].eq(sink.be[4* 4:]),
                source.be[4* 8:4* 9].eq(sink.be[4* 5:]),
                source.be[4* 9:4*10].eq(sink.be[4* 6:]),
                source.be[4*10:4*11].eq(sink.be[4* 7:]),
                source.be[4*11:4*12].eq(sink.be[4* 8:]),
                source.be[4*12:4*13].eq(sink.be[4* 9:]),
                source.be[4*13:4*14].eq(sink.be[4*10:]),
                source.be[4*14:4*15].eq(sink.be[4*11:]),
                source.be[4*15:4*16].eq(sink.be[4*12:]),
            ),
            If(source.valid & source.ready,
                sink.ready.eq(~last),
                If(source.last,
                    NextState("HEADER")
                )
            )
        )

class LitePCIeTLPHeaderInserter512b4DWs(Module):
    def __init__(self):
        self.sink   = sink   = stream.Endpoint(tlp_raw_layout(512))
        self.source = source = stream.Endpoint(phy_layout(512))

        # # #

        dat  = Signal(512,    reset_less=True)
        be   = Signal(512//8, reset_less=True)
        last = Signal(        reset_less=True)
        self.sync += [
            If(sink.valid & sink.ready,
                dat.eq(sink.dat),
                be.eq(sink.be),
                last.eq(sink.last)
            )
        ]

        self.submodules.fsm = fsm = FSM(reset_state="HEADER")
        fsm.act("HEADER",
            sink.ready.eq(1),
            If(sink.valid & sink.first,
                sink.ready.eq(0),
                source.valid.eq(1),
                source.first.eq(1),
                source.last.eq(sink.last & (sink.be[4*12:] == 0)),

                source.dat[32* 0:32* 1].eq(sink.header[32*0:]),
                source.dat[32* 1:32* 2].eq(sink.header[32*1:]),
                source.dat[32* 2:32* 3].eq(sink.header[32*2:]),
                source.dat[32* 3:32* 4].eq(sink.header[32*3:]),
                source.dat[32* 4:32* 5].eq(sink.dat[32* 0:]),
                source.dat[32* 5:32* 6].eq(sink.dat[32* 1:]),
                source.dat[32* 6:32* 7].eq(sink.dat[32* 2:]),
                source.dat[32* 7:32* 8].eq(sink.dat[32* 3:]),
                source.dat[32* 8:32* 9].eq(sink.dat[32* 4:]),
                source.dat[32* 9:32*10].eq(sink.dat[32* 5:]),
                source.dat[32*10:32*11].eq(sink.dat[32* 6:]),
                source.dat[32*11:32*12].eq(sink.dat[32* 7:]),
                source.dat[32*12:32*13].eq(sink.dat[32* 8:]),
                source.dat[32*13:32*14].eq(sink.dat[32* 9:]),
                source.dat[32*14:32*15].eq(sink.dat[32*10:]),
                source.dat[32*15:32*16].eq(sink.dat[32*11:]),

                source.be[4* 0:4* 1].eq(0xf),
                source.be[4* 1:4* 2].eq(0xf),
                source.be[4* 2:4* 3].eq(0xf),
                source.be[4* 3:4* 4].eq(0xf),
                source.be[4* 4:4* 5].eq(sink.be[4* 0:]),
                source.be[4* 5:4* 6].eq(sink.be[4* 1:]),
                source.be[4* 6:4* 7].eq(sink.be[4* 2:]),
                source.be[4* 7:4* 8].eq(sink.be[4* 3:]),
                source.be[4* 8:4* 9].eq(sink.be[4* 4:]),
                source.be[4* 9:4*10].eq(sink.be[4* 5:]),
                source.be[4*10:4*11].eq(sink.be[4* 6:]),
                source.be[4*11:4*12].eq(sink.be[4* 7:]),
                source.be[4*12:4*13].eq(sink.be[4* 8:]),
                source.be[4*13:4*14].eq(sink.be[4* 9:]),
                source.be[4*14:4*15].eq(sink.be[4*10:]),
                source.be[4*15:4*16].eq(sink.be[4*11:]),

                If(source.valid & source.ready,
                    sink.ready.eq(1),
                    If(~source.last,
                        NextState("DATA"),
                    )
                )
            )
        )
        fsm.act("DATA",
            source.valid.eq(sink.valid | last),
            source.last.eq(last),

            source.dat[32* 0:32* 1].eq(     dat[32*12:]),
            source.dat[32* 1:32* 2].eq(     dat[32*13:]),
            source.dat[32* 2:32* 3].eq(     dat[32*14:]),
            source.dat[32* 3:32* 4].eq(     dat[32*15:]),
            source.dat[32* 4:32* 5].eq(sink.dat[32* 0:]),
            source.dat[32* 5:32* 6].eq(sink.dat[32* 1:]),
            source.dat[32* 6:32* 7].eq(sink.dat[32* 2:]),
            source.dat[32* 7:32* 8].eq(sink.dat[32* 3:]),
            source.dat[32* 8:32* 9].eq(sink.dat[32* 4:]),
            source.dat[32* 9:32*10].eq(sink.dat[32* 5:]),
            source.dat[32*10:32*11].eq(sink.dat[32* 6:]),
            source.dat[32*11:32*12].eq(sink.dat[32* 7:]),
            source.dat[32*12:32*13].eq(sink.dat[32* 8:]),
            source.dat[32*13:32*14].eq(sink.dat[32* 9:]),
            source.dat[32*14:32*15].eq(sink.dat[32*10:]),
            source.dat[32*15:32*16].eq(sink.dat[32*11:]),

            source.be[4*0:4*1].eq(be[4*12:]),
            source.be[4*1:4*2].eq(be[4*13:]),
            source.be[4*2:4*3].eq(be[4*14:]),
            source.be[4*3:4*4].eq(be[4*15:]),

            If(last,
                source.be[4*4:4*16].eq(0x0)
            ).Else(
                source.be[ 4*4: 4*5].eq(sink.be[4* 0:]),
                source.be[ 4*5: 4*6].eq(sink.be[4* 1:]),
                source.be[ 4*6: 4*7].eq(sink.be[4* 2:]),
                source.be[ 4*7: 4*8].eq(sink.be[4* 3:]),
                source.be[ 4*8: 4*9].eq(sink.be[4* 4:]),
                source.be[ 4*9:4*10].eq(sink.be[4* 5:]),
                source.be[4*10:4*11].eq(sink.be[4* 6:]),
                source.be[4*11:4*12].eq(sink.be[4* 7:]),
                source.be[4*12:4*13].eq(sink.be[4* 8:]),
                source.be[4*13:4*14].eq(sink.be[4* 9:]),
                source.be[4*14:4*15].eq(sink.be[4*10:]),
                source.be[4*15:4*16].eq(sink.be[4*11:]),
            ),
            If(source.valid & source.ready,
                sink.ready.eq(~last),
                If(source.last,
                    NextState("HEADER")
                )
            )
        )

class LitePCIeTLPHeaderInserter512b(LitePCIeTLPHeaderInserter3DWs4DWs):
    def __init__(self, fmt):
        LitePCIeTLPHeaderInserter3DWs4DWs.__init__(self,
            data_width               = 512,
            header_inserter_3dws_cls = LitePCIeTLPHeaderInserter512b3DWs,
            header_inserter_4dws_cls = LitePCIeTLPHeaderInserter512b4DWs,
            fmt                      = fmt,
        )

# LitePCIeTLPPacketizer ----------------------------------------------------------------------------

class LitePCIeTLPPacketizer(Module):
    def __init__(self, data_width, endianness, address_width=32):
        assert data_width%32 == 0
        assert address_width in [32, 64]
        if address_width == 64:
            assert data_width in [64, 128, 256, 512]
        self.req_sink = req_sink = stream.Endpoint(request_layout(data_width, address_width))
        self.cmp_sink = cmp_sink = stream.Endpoint(completion_layout(data_width))
        self.source   = stream.Endpoint(phy_layout(data_width))

        # # #

        # Format TLP request and encode it ---------------------------------------------------------
        self.tlp_req = tlp_req = stream.Endpoint(tlp_request_layout(data_width))
        self.comb += [
            tlp_req.valid.eq(req_sink.valid),
            req_sink.ready.eq(tlp_req.ready),
            tlp_req.first.eq(req_sink.first),
            tlp_req.last.eq(req_sink.last),

            tlp_req.type.eq(0b00000),
            If(req_sink.we,
                tlp_req.fmt.eq( fmt_dict[ f"mem_wr32"]),
            ).Else(
                tlp_req.fmt.eq( fmt_dict[ f"mem_rd32"]),
            ),
            tlp_req.address.eq(req_sink.adr),
        ]
        if address_width == 64:
            self.comb += [
                # Use WR64/RD64 only when 64-bit Address's MSB != 0, else use WR32/RD32.
                If(req_sink.adr[32:] != 0,
                    # Address's MSB on DW2, LSB on DW3 with 64-bit addressing: Requires swap due to
                    # Packetizer's behavior.
                    tlp_req.address[:32].eq(req_sink.adr[32:]),
                    tlp_req.address[32:].eq(req_sink.adr[:32]),
                    If(req_sink.we,
                        tlp_req.fmt.eq( fmt_dict[ f"mem_wr64"]),
                    ).Else(
                        tlp_req.fmt.eq( fmt_dict[ f"mem_rd64"]),
                    ),
                )
            ]

        self.comb += [
            tlp_req.tc.eq(0),
            tlp_req.td.eq(0),
            tlp_req.ep.eq(0),
            tlp_req.attr.eq(0),
            tlp_req.length.eq(req_sink.len),

            tlp_req.requester_id.eq(req_sink.req_id),
            tlp_req.tag.eq(req_sink.tag),
            If(req_sink.len > 1,
                tlp_req.last_be.eq(0xf)
            ).Else(
                tlp_req.last_be.eq(0x0)
            ),
            tlp_req.first_be.eq(0xf),
            tlp_req.dat.eq(req_sink.dat),
            If(req_sink.we,
                tlp_req.be.eq(2**(data_width//8)-1)
            ).Else(
                tlp_req.be.eq(0x00)
            )
        ]

        tlp_raw_req        = stream.Endpoint(tlp_raw_layout(data_width))
        tlp_raw_req_header = Signal(len(tlp_raw_req.header))
        self.comb += [
            tlp_req.connect(tlp_raw_req, omit={*tlp_request_header_fields.keys()}),
            tlp_raw_req.fmt.eq(tlp_req.fmt),
            tlp_request_header.encode(tlp_req, tlp_raw_req_header),
        ]
        self.comb += dword_endianness_swap(
            src        = tlp_raw_req_header,
            dst        = tlp_raw_req.header,
            data_width = data_width,
            endianness = endianness,
            mode       = "dat",
            ndwords    = 4
        )

        # Format TLP completion and encode it ------------------------------------------------------
        self.tlp_cmp = tlp_cmp = stream.Endpoint(tlp_completion_layout(data_width))
        self.comb += [
            tlp_cmp.valid.eq(cmp_sink.valid),
            cmp_sink.ready.eq(tlp_cmp.ready),
            tlp_cmp.first.eq(cmp_sink.first),
            tlp_cmp.last.eq(cmp_sink.last),

            tlp_cmp.tc.eq(0),
            tlp_cmp.td.eq(0),
            tlp_cmp.ep.eq(0),
            tlp_cmp.attr.eq(0),
            tlp_cmp.length.eq(cmp_sink.len),

            tlp_cmp.completer_id.eq(cmp_sink.cmp_id),
            If(cmp_sink.err,
                tlp_cmp.type.eq(type_dict["cpl"]),
                tlp_cmp.fmt.eq( fmt_dict["cpl"]),
                tlp_cmp.status.eq(cpl_dict["ur"])
            ).Else(
                tlp_cmp.type.eq(type_dict["cpld"]),
                tlp_cmp.fmt.eq( fmt_dict["cpld"]),
                tlp_cmp.status.eq(cpl_dict["sc"])
            ),
            tlp_cmp.bcm.eq(0),
            tlp_cmp.byte_count.eq(cmp_sink.len*4),

            tlp_cmp.requester_id.eq(cmp_sink.req_id),
            tlp_cmp.tag.eq(cmp_sink.tag),
            tlp_cmp.lower_address.eq(cmp_sink.adr),

            tlp_cmp.dat.eq(cmp_sink.dat),
            If(cmp_sink.last & cmp_sink.first,
                tlp_cmp.be.eq(0xf)
            ).Else(
                tlp_cmp.be.eq(2**(data_width//8)-1)
            ),
        ]

        tlp_raw_cmp        = stream.Endpoint(tlp_raw_layout(data_width))
        tlp_raw_cmp_header = Signal(len(tlp_raw_cmp.header))
        self.comb += [
            tlp_cmp.connect(tlp_raw_cmp, omit={*tlp_completion_header_fields.keys()}),
            tlp_raw_cmp.fmt.eq(tlp_cmp.fmt),
            tlp_completion_header.encode(tlp_cmp, tlp_raw_cmp_header),
        ]
        self.comb += dword_endianness_swap(
            src        = tlp_raw_cmp_header,
            dst        = tlp_raw_cmp.header,
            data_width = data_width,
            endianness = endianness,
            mode       = "dat",
            ndwords    = 4
        )

        # Arbitrate --------------------------------------------------------------------------------

        tlp_raw = stream.Endpoint(tlp_raw_layout(data_width))
        self.submodules.arbitrer = Arbiter(
            masters = [tlp_raw_req, tlp_raw_cmp],
            slave   = tlp_raw
        )

        # Insert header ----------------------------------------------------------------------------
        header_inserter_cls = {
            64 : LitePCIeTLPHeaderInserter64b,
           128 : LitePCIeTLPHeaderInserter128b,
           256 : LitePCIeTLPHeaderInserter256b,
           512 : LitePCIeTLPHeaderInserter512b,
        }
        header_inserter = header_inserter_cls[data_width](fmt=tlp_raw.fmt)
        self.submodules += header_inserter
        self.comb += tlp_raw.connect(header_inserter.sink)
        self.comb += header_inserter.source.connect(self.source, omit={"data", "be"})
        for name in ["dat", "be"]:
            self.comb += dword_endianness_swap(
                src        = getattr(header_inserter.source, name),
                dst        = getattr(self.source, name),
                data_width = data_width,
                endianness = endianness,
                mode       = name,
            )



#
# This file is part of LitePCIe.
#
# Copyright (c) 2015-2021 Florent Kermarrec <florent@enjoy-digital.fr>
# SPDX-License-Identifier: BSD-2-Clause

from migen import *

from litepcie.common import *
from litepcie.core.common import *
from litepcie.tlp.common import *


class LitePCIeTLPController(Module):
    """LitePCIe TLP requests/completions controller.

    Arbitrate/throttle TLP requests and reorder/assemble/redirect completions.
    """
    def __init__(self, data_width, address_width, max_pending_requests, cmp_bufs_buffered=True):
        self.master_in  = LitePCIeMasterInternalPort(data_width, address_width)
        self.master_out = LitePCIeMasterInternalPort(data_width, address_width)

        # # #

        req_sink, req_source = self.master_in.sink,    self.master_out.sink
        cmp_sink, cmp_source = self.master_out.source, self.master_in.source

        # Parameters.
        tag_bits = log2_int(max_pending_requests)

        # Tag queue --------------------------------------------------------------------------------
        # The tag queue is filled initially with the tags that will be used to issue read requests
        # to the host. A tag is dequeued when a read requests is issued to the host and queued when
        # a readcomplementation is received from the host.
        tag_queue = SyncFIFO([("tag", tag_bits)], max_pending_requests, buffered=True)
        self.submodules.tag_queue = tag_queue

        # Requests queue ---------------------------------------------------------------------------
        # Store the read requests tags as emitted to the host, datas will be dequeued in this order
        req_queue_layout = [("tag", tag_bits), ("channel", 8), ("user_id", 8)]
        req_queue        = SyncFIFO(req_queue_layout, max_pending_requests, buffered=True)
        self.submodules.req_queue = req_queue

        # Requests Management ----------------------------------------------------------------------

        # Connect Data-Path.
        self.comb += req_sink.connect(req_source, omit={"valid", "ready", "tag"})
        self.submodules.req_fsm = req_fsm = FSM(reset_state="WAIT-REQ")

        # FSM.
        req_fsm.act("WAIT-REQ",
            # Wait for a TLP Request...
            If(req_sink.valid & req_sink.first,
                # TLP Write: We can send the request directly.
                If(req_sink.we,
                   NextState("SEND-WRITE-REQ")
                # TLP Read:  We can send the request when one tag available and space in req_queue.
                ).Elif(tag_queue.source.valid & req_queue.sink.ready,
                   NextState("SEND-READ-REQ")
                )
            )
        )
        req_fsm.act("SEND-WRITE-REQ",
            # Connect Control-Path.
            req_sink.connect(req_source, keep={"valid", "ready"}),
            req_source.tag.eq(32),
            # End Request and return to Wait on last valid cycle.
            If(req_source.valid & req_source.ready & req_source.last,
                NextState("WAIT-REQ")
            )
        )
        req_fsm.act("SEND-READ-REQ",
            # Connect Control-Path.
            req_sink.connect(req_source, keep={"valid", "ready"}),
            req_source.tag.eq(tag_queue.source.tag),
            # End Request and return to Wait on last valid cycle.
            If(req_source.valid & req_source.ready & req_source.last,
                # Pop Tag from tag_queue.
                tag_queue.source.ready.eq(1),
                # Push Req to req_queue.
                req_queue.sink.valid.eq(1),
                req_queue.sink.tag.eq(tag_queue.source.tag),
                req_sink.connect(req_queue.sink, keep={"channel", "user_id"}),
                NextState("WAIT-REQ")
            )
        )

        # Completions Data Buffers (Reordering) ----------------------------------------------------
        #                          ┌───────────────┐
        #   Read Requests'Tags ────►   Req Queue   ├───┐
        #                          └───────────────┘   │
        #                                              │
        #                                              ▼
        #                 ┌─► Demux                   Mux
        #             Tag │   ┌──┐    ┌──────────┐   ┌───┐
        #                 │   │  ├────► Buffer 0 ├───►   │
        #                 │   │  │    └──────────┘   │   │
        #                 │   │  │                   │   │
        #                 │   │  │    ┌──────────┐   │   │
        #        Cmp In ──┴───►  ├────► Buffer . ├───►   ├─────► Cmp Out
        #                     │  │    └──────────┘   │   │
        #                     │  │                   │   │
        #                     │  │    ┌──────────┐   │   │
        #                     │  ├────► Buffer N ├───►   │
        #                     └──┘    └──────────┘   └───┘
        #
        #

        cmp_reorder = stream.Endpoint(completion_layout(data_width))
        cmp_bufs    = []

        # Create Buffers.
        for i in range(max_pending_requests):
            cmp_buf_depth = 4*max_request_size//(data_width//8)
            cmp_buf       = SyncFIFO(completion_layout(data_width), cmp_buf_depth, buffered=cmp_bufs_buffered)
            cmp_bufs.append(cmp_buf)
        self.submodules += cmp_bufs

        # Connect Cmp Input to Buffers (based on incoming Tag).
        self.comb += Case(cmp_reorder.tag,
            {i: cmp_reorder.connect(cmp_bufs[i].sink) for i in range(len(cmp_bufs))})

        # Connect Buffers to Cmp Output (based on Tag from req_ueue).
        self.comb += Case(req_queue.source.tag,
            {i: cmp_bufs[i].source.connect(cmp_source) for i in range(len(cmp_bufs))})
        self.comb += [
            # Pop Req from req_queue when Req is fully received.
            If(cmp_source.valid & cmp_source.last & cmp_source.end,
                req_queue.source.ready.eq(cmp_source.ready)
            ),
            req_queue.source.connect(cmp_source, keep={"channel", "user_id"}),
        ]

        # Completions Management -------------------------------------------------------------------

        fill_tag = Signal(tag_bits)

        # Connect Data-Path.
        self.comb += cmp_sink.connect(cmp_reorder, omit={"valid", "ready"})

        # FSM
        self.submodules.cmp_fsm = cmp_fsm = FSM(reset_state="FILL-TAG-QUEUE")
        cmp_fsm.act("FILL-TAG-QUEUE",
            # Pre-fill Tags.
            tag_queue.sink.valid.eq(1),
            tag_queue.sink.tag.eq(fill_tag),
            NextValue(fill_tag, fill_tag + 1),
            If(fill_tag == (max_pending_requests - 1),
                NextState("WAIT")
            )
        )
        cmp_fsm.act("WAIT",
            # Wait for a TLP Completion...
            If(cmp_sink.valid & cmp_sink.first,
                NextState("RUN")
            ).Else(
                cmp_sink.ready.eq(1)
            )
        )
        cmp_fsm.act("RUN",
            # Connect Control-Path.
            cmp_sink.connect(cmp_reorder, keep={"valid", "ready"}),
            # Push incoming Tag to tag_queue when Cmp is fully received.
            If(cmp_sink.valid & cmp_sink.ready & cmp_sink.last,
                If(cmp_sink.end,
                    tag_queue.sink.valid.eq(1),
                    tag_queue.sink.tag.eq(cmp_sink.tag)
                ),
                NextState("WAIT")
            )
        )



#
# This file is part of LitePCIe.
#
# Copyright (c) 2015-2022 Florent Kermarrec <florent@enjoy-digital.fr>
# Copyright (c) 2020 Antmicro <www.antmicro.com>
# SPDX-License-Identifier: BSD-2-Clause

from migen import *

from litex.soc.interconnect import stream
from litex.soc.interconnect.csr import *

from litepcie.common import *
from litepcie.tlp.common import *

# Constants/Layouts --------------------------------------------------------------------------------

def descriptor_layout(address_width=32, with_user_id=False):
    layout = [("address", address_width), ("length",  24), ("irq_disable", 1), ("last_disable", 1)]
    if with_user_id:
        layout += [("user_id", 8)]
    return EndpointDescription(layout)


# LitePCIeDMAScatterGather --------------------------------------------------------------------------

class LitePCIeDMAScatterGather(Module, AutoCSR):
    """LitePCIe DMA Scatter-Gather

    Software programmable table storing a list of DMA descriptors.

                               Mode
                                │
                              ┌─▼─┐  ┌───────────────────┐
       Descriptor to program  │   │  │                   │
             (Prog Mode) ─────►   │  │                   │
                              │   │  │    ScatterGather  │
                              │   ├──►     Table(FIFO)   ├──┬───► Descriptor (To DMA).
                          ┌───►   │  │                   │  │
                          │   │   │  │                   │  │
                          │   └───┘  └───────────────────┘  │
                          │                                 │
                          └─────────────────────────────────┘
                                    Refill (Loop Mode)


    A DMA descriptor is composed of:
    - a 32/64-bit address: The base address of the Host where the data stream should be written/read.
    - a 24-bit length : The length of the data stream (bytes).
    - a 8-bit control : Dynamic controls (ex: Disable IRQ generation, disable Last handling).

    The table is implemented as a FIFO initially filled by software. Once enabled, the DMA gets the
    descriptors from this table and executes them.

    This module has two modes:
    - Prog mode: Used to program the table by software and for cases where automatic refill of the
    table is not needed: A descriptor is only executed once and when all the descriptors have been
    executed (ie the table is empty), the DMA just stops until the next software refill.
    - Loop mode: Used once the table has been filled by software in PROG mode and allow continuous
    Scatter-Gather DMA: Each descriptor sent to the DMA is refilled to the table.

    In Loop mode, a loop status is maintained by the hardware for the software synchronization of the
    DMA buffers. (Even if a MSI IRQ is generated after a descriptor has been executed, since IRQ can
    potentially be lost, it's safer for the software to just use the hardware loop status than to
    maintain a software loop status based MSI IRQ reception).
    """
    def __init__(self, depth, address_width=32):
        assert address_width in [32, 64]
        # Stream Endpoint.
        self.source = source = stream.Endpoint(descriptor_layout(address_width=address_width))

        # Control/Status.
        self.value = CSRStorage(64, reset_less=True, fields=[
            CSRField("address_lsb",  size=32, description="32-bit LSB Address of the descriptor (bytes-aligned)."),
            CSRField("length",       size=24, description="24-bit Length  of the descriptor (in bytes)."),
            CSRField("irq_disable",  size=1,  description="IRQ Disable Control of the descriptor."),
            CSRField("last_disable", size=1,  description="Last Disable Control of the descriptor.")
            ], description="64-bit DMA descriptor to be written to the table.")
        self.we = CSRStorage(32, description="Write and 32-bit MSB Address of the descriptor (bytes-aligned)", fields=[
            CSRField("address_msb", size=32, description="32-bit MSB Address of the descriptor (bytes-aligned), in 64-bit mode."),
        ])
        self.loop_prog_n = CSRStorage(description="""Mode Selection.\n
            ``0``: **Prog** mode / ``1``: **Loop** mode.\n
            **Prog** mode should be used to program the table by software and for cases where automatic
            refill of the table is not needed: A descriptor is only executed once and when all the
            descriptors have been executed (ie the table is empty), the DMA just stops until the next
            software refill.\n
            **Loop** mode should be used once the table has been filled by software in **Prog** mode
            and allow continuous Scatter-Gather DMA: Each descriptor sent to the DMA is refilled to the table.
            """)
        self.loop_status = CSRStatus(fields=[
            CSRField("index", size=16, description= "Index of the last descriptor executed in the DMA descriptor table."),
            CSRField("count", size=16, description= "Loops of the DMA descriptor table since started."),
            ], description="Loop monitoring for software synchronization.")
        self.level = CSRStatus(bits_for(depth), description="Number descriptors in the table.")
        self.reset = CSRStorage(description="A write to this register resets the table.")

        # # #

        # Table (FIFO) -----------------------------------------------------------------------------
        table = stream.SyncFIFO(descriptor_layout(address_width=address_width), depth)
        table = ResetInserter()(table)
        self.submodules += table
        self.comb += table.reset.eq(self.reset.storage & self.reset.re)
        self.comb += self.level.status.eq(table.level)

        # Table Write logic ------------------------------------------------------------------------
        def table_sink_address_map():
            address_map = []
            address_map.append(table.sink.address[0:32].eq(self.value.fields.address_lsb))
            if address_width == 64:
                address_map.append(table.sink.address[32:64].eq(self.we.fields.address_msb))
            return address_map

        prog_mode = (self.loop_prog_n.storage == 0)
        loop_mode = (self.loop_prog_n.storage == 1)
        self.sync += [
            # In Prog mode, the Table is filled through the CSRs.
            If(prog_mode,
                *table_sink_address_map(),
                table.sink.length.eq(self.value.fields.length),
                table.sink.irq_disable.eq(self.value.fields.irq_disable),
                table.sink.last_disable.eq(self.value.fields.last_disable),
                table.sink.first.eq(table.level == 0),
                table.sink.valid.eq(self.we.re),
            # In Loop mode, the Table is automatically refilled.
            ).Else(
                table.source.connect(table.sink, omit={"valid", "ready"}),
                table.sink.valid.eq(table.source.valid & table.source.ready),
            )
        ]

        # Table Read logic -------------------------------------------------------------------------
        self.comb += table.source.connect(source)

        # Loop Status (For Software Sychronization in Loop mode) -----------------------------------
        loop_first = Signal()
        loop_index = self.loop_status.fields.index
        loop_count = self.loop_status.fields.count
        self.sync += [
            # Reset Loop Index/Count on Table reset.
            If(table.reset,
                loop_first.eq(1),
                loop_index.eq(0),
                loop_count.eq(0),
            # When a Descriptor is consumned...
            ).Elif(table.source.valid & table.source.ready,
                # Update Loop Status with current Loop Index/Count.
                # Loop Mode.
                If(loop_mode & table.source.first,
                    # Reset Index.
                    loop_index.eq(0),
                    # Increment Count (except on first since we want (index, count) == (0,0)).
                    loop_first.eq(0),
                    loop_count.eq(loop_count + Cat(~loop_first)),
                # Prog Mode.
                ).Else(
                    # Increment Index.
                    loop_index.eq(loop_index + 1),
                    # Increment Count.
                    If(loop_index == (2**16-1),
                        loop_count.eq(loop_count + 1)
                    )
                )
            )
        ]

# LitePCIeDMADescriptorSplitter --------------------------------------------------------------------

class LitePCIeDMADescriptorSplitter(Module, AutoCSR):
    """LitePCIe DMA Descriptor Splitter

    Splits descriptors from LitePCIeDMAScatterGather in shorter descriptors of:
    - Maximum Payload Size for Writes.
    - Maximum Request Size for Reads.

    Descriptors from LitePCIeDMAScatterGather have a maximum length of 16Mb (24-bits). It is not
    possible to do such long Writes/Reads on the PCIe bus. At the PCIe enumeration, Maximum Payload
    and Request Sizes are negotiated between the Host and the Device. Writes are limited to Maximum
    Payload Size, Reads are limited to Maximum Request Size. Each descriptor is then split in
    several shorter descriptors.
    """
    def __init__(self, max_size, address_width):
        # Stream Endpoints.
        self.sink   =   sink = stream.Endpoint(descriptor_layout(address_width=address_width))
        self.source = source = stream.Endpoint(descriptor_layout(address_width=address_width, with_user_id=True))

        self.terminate = Signal() # Early Termination.

        # # #

        desc_length  = Signal(32)
        desc_offset  = Signal(32)
        desc_id      = Signal(32)

        # FSM --------------------------------------------------------------------------------------
        self.submodules.fsm = fsm = FSM(reset_state="IDLE")
        fsm.act("IDLE",
            # Set Descriptor Offset/Length.
            NextValue(desc_offset, 0),
            NextValue(desc_length, sink.length),
            # Wait for a Descriptor and go to Run.
            If(sink.valid,
                NextState("SPLIT")
            )
        )
        # Split Data-Path.
        self.comb += [
            source.address.eq(sink.address + desc_offset),
            source.irq_disable.eq(sink.irq_disable),
            source.last_disable.eq(sink.last_disable),
            source.user_id.eq(desc_id),
        ]
        fsm.act("SPLIT",
            # Split Control-Path.
            source.valid.eq(1),
            source.first.eq(desc_offset == 0),
            # Full Descriptor when Length > max_size.
            If(desc_length > max_size,
                source.last.eq(self.terminate),
                source.length.eq(max_size),
            # Partial Descriptor when Length <= max_size.
            ).Else(
                source.last.eq(1),
                source.length.eq(desc_length),
            ),
            # When Descriptor is accepted...
            If(source.ready,
                # Increment Offset.
                NextValue(desc_offset, desc_offset + max_size),
                # Decrement Length.
                NextValue(desc_length, desc_length - max_size),
                # When Last....
                If(source.last,
                    # Accept Descriptor.
                    sink.ready.eq(1),
                    # Increment ID.
                    NextValue(desc_id, desc_id + 1),
                    # Go to Idle.
                    NextState("IDLE")
                )
            )
        )

# LitePCIeDMAReader --------------------------------------------------------------------------------

class LitePCIeDMAReader(Module, AutoCSR):
    """LitePCIe DMA Reader

    Generates a data stream from Host's memory.

    This module allows Scatter-Gather DMAs from Host's memory to data stream in the FPGA. The DMA
    descriptors, stored in a software programmable table, are split and executed as Read Requests
    on the PCIe bus.

    A Read Request is only sent to the Host when enough space is available in the Data FIFO to store
    the requested data.

    A MSI IRQ can be generated when a descriptor has been executed.
    """
    def __init__(self, endpoint, port, with_table=True, table_depth=256, address_width=32, with_splitter_buffer=True):
        self.port = port
        # Stream Endpoint.
        self.source = stream.Endpoint(dma_layout(endpoint.phy.data_width))

        # Control.
        self._enable = CSRStorage(size=2, description="DMA Reader Control. Write ``1`` to enable DMA Reader.", reset=0 if with_table else 1)

        # IRQ.
        self.irq = Signal()

        # # #

        # CSR/Parameters ---------------------------------------------------------------------------
        self.enable = enable = self._enable.storage[0]

        length_shift          = log2_int(endpoint.phy.data_width//8)
        max_words_per_request = max_request_size//(endpoint.phy.data_width//8)
        max_pending_words     = endpoint.max_pending_requests*max_words_per_request

        # Table ------------------------------------------------------------------------------------
        if with_table:
            self.submodules.table = LitePCIeDMAScatterGather(table_depth, address_width=address_width)
        else:
            self.desc_sink = stream.Endpoint(descriptor_layout(address_width=address_width)) # Expose a Descriptor sink.

        # Splitter ---------------------------------------------------------------------------------
        # DMA descriptors need to be splitted in descriptors of max_request_size (negotiated at link-up)
        splitter = LitePCIeDMADescriptorSplitter(
            max_size      = endpoint.phy.max_request_size,
            address_width = address_width
        )
        splitter = ResetInserter()(splitter)
        if with_splitter_buffer: # For timings.
            splitter = BufferizeEndpoints({"source": DIR_SOURCE})(splitter)
        self.submodules.splitter = splitter
        if with_table:
            self.comb += self.table.source.connect(splitter.sink)
        else:
            self.comb += self.desc_sink.connect(splitter.sink)

        # User ID ----------------------------------------------------------------------------------
        last_user_id = Signal(8, reset=255)
        self.sync += If(port.sink.valid & port.sink.first & port.sink.ready,
            last_user_id.eq(port.sink.user_id)
        )

        # Data FIFO --------------------------------------------------------------------------------
        data_fifo_depth = 4*max_pending_words
        data_fifo = SyncFIFO(dma_layout(endpoint.phy.data_width), data_fifo_depth, buffered=True)
        self.submodules.data_fifo = ResetInserter()(data_fifo)
        self.comb += [
            # Connect Data FIFO to Source.
            data_fifo.source.connect(self.source),
            # When Enabled, connect Sink to Data FIFO.
            If(enable,
                port.sink.connect(data_fifo.sink, keep={"valid", "ready"}),
                data_fifo.sink.data.eq(port.sink.dat),
                data_fifo.sink.first.eq(port.sink.first & (port.sink.user_id != last_user_id)),
            # Else accept incoming Port Data.
            ).Else(
                port.sink.ready.eq(1)
            )
        ]

        # Pending words ----------------------------------------------------------------------------
        pending_words         = Signal(max=data_fifo_depth + 1)
        pending_words_queue   = Signal.like(pending_words)
        pending_words_dequeue = Signal.like(pending_words)
        self.comb += [
            # Queue Pending words as Read Requests are emitted.
            If(splitter.source.valid & splitter.source.ready,
                pending_words_queue.eq(splitter.source.length[length_shift:])
            ),
            # Dequeue Pending words as Read Responses are received.
            If(data_fifo.source.valid & data_fifo.source.ready,
                pending_words_dequeue.eq(1)
            ),
        ]
        # Update Pending words.
        self.sync += pending_words.eq(pending_words + pending_words_queue - pending_words_dequeue)
        self.sync += If(~enable, pending_words.eq(0))

        # FSM --------------------------------------------------------------------------------------
        self.submodules.fsm = fsm = FSM(reset_state="IDLE")
        fsm.act("IDLE",
            # Reset Splitter/FIFO when disabled.
            If(~enable,
                splitter.reset.eq(1),
                data_fifo.reset.eq(1),
            # Else wait for a Descriptor and to have enough Space to generate the Request.
            ).Elif(splitter.source.valid & (pending_words < (data_fifo_depth - max_words_per_request)),
                NextState("MEM-RD-REQ"),
            )
        )
        # Report Idle Status.
        self.sync += self._enable.storage[1].eq(fsm.ongoing("IDLE"))
        # Request Data-Path.
        self.comb += [
            port.source.channel.eq(port.channel),
            port.source.user_id.eq(splitter.source.user_id),
            port.source.first.eq(1),
            port.source.last.eq(1),
            port.source.we.eq(0),
            port.source.adr.eq(splitter.source.address),
            port.source.len.eq(splitter.source.length[2:]),
            port.source.req_id.eq(endpoint.phy.id),
            port.source.dat.eq(0),
        ]
        fsm.act("MEM-RD-REQ",
            # Request Control-Path.
            port.source.valid.eq(1),
            # When Request is accepted...
            If(port.source.ready,
                # Accept Descriptor.
                splitter.source.ready.eq(1),
                # Return to Idle.
                NextState("IDLE"),
            )
        )

        # IRQ --------------------------------------------------------------------------------------
        self.comb += If(splitter.source.valid & splitter.source.ready & splitter.source.last,
            self.irq.eq(~splitter.source.irq_disable)
        )

# LitePCIeDMAWriter --------------------------------------------------------------------------------

class LitePCIeDMAWriter(Module, AutoCSR):
    """LitePCIe DMA Writer

    Stores a data stream to Host's memory.

    This module allows Scatter-Gather DMAs from a data stream in the FPGA to Host's memory. The DMA
    descriptors, stored in a software programmable table, are split and executed as Write Requests
    on the PCIe bus.

    A Write Request is only sent to the Host when enough data are available for the current split
    descriptor.

    A MSI IRQ can be generated when a descriptor has been executed.
    """
    def __init__(self, endpoint, port, with_table=True, table_depth=256, address_width=32, with_splitter_buffer=True):
        self.port = port
        # Stream Endpoint.
        self.sink = sink = stream.Endpoint(dma_layout(endpoint.phy.data_width))

        # Control.
        self._enable = CSRStorage(size=2, description="DMA Writer Control. Write ``1`` to enable DMA Writer.", reset=0 if with_table else 1)

        # IRQ.
        self.irq = Signal()

        # # #

        # CSR/Parameters ---------------------------------------------------------------------------
        self.enable = enable = self._enable.storage[0]

        length_shift          = log2_int(endpoint.phy.data_width//8)
        max_words_per_request = max_payload_size//(endpoint.phy.data_width//8)

        # Table ------------------------------------------------------------------------------------
        if with_table:
            self.submodules.table = LitePCIeDMAScatterGather(table_depth, address_width)
        else:
            self.desc_sink = stream.Endpoint(descriptor_layout(address_width=address_width)) # Expose a Descriptor sink.

        # Splitter ---------------------------------------------------------------------------------
        # DMA descriptors need to be splitted in descriptors of max_request_size (negotiated at link-up)
        splitter = LitePCIeDMADescriptorSplitter(
            max_size      = endpoint.phy.max_payload_size,
            address_width = address_width
        )
        splitter = ResetInserter()(splitter)
        if with_splitter_buffer: # For timings.
            # FIXME: Prevent early termination, so don't use when early termination is required.
            splitter = BufferizeEndpoints({"source": DIR_SOURCE})(splitter)
        self.submodules.splitter = splitter
        if with_table:
            self.comb += self.table.source.connect(splitter.sink)
        else:
            self.comb += self.desc_sink.connect(splitter.sink)

        # Data FIFO --------------------------------------------------------------------------------
        data_fifo_depth = 4*max_words_per_request
        data_fifo = stream.SyncFIFO([("data", endpoint.phy.data_width)], data_fifo_depth, buffered=True)
        self.submodules.data_fifo = ResetInserter()(data_fifo)
        # By default, accept incoming stream when disabled.
        self.comb += sink.ready.eq(1)
        # When Enabled, connect Sink to Data FIFO.
        self.comb += If(enable, sink.connect(data_fifo.sink))

        # FSM --------------------------------------------------------------------------------------
        req_count = Signal.like(splitter.source.length)
        self.submodules.fsm = fsm = FSM(reset_state="IDLE")
        fsm.act("IDLE",
            # Reset Request Count.
            NextValue(req_count, 0),
            # Reset Splitter/FIFO when disabled.
            If(~enable,
                splitter.reset.eq(1),
                data_fifo.reset.eq(1),
            # Else wait for a Descriptor and to have enough Data to generate the Request.
            ).Elif(splitter.source.valid & (data_fifo.level >= splitter.source.length[length_shift:]),
                NextState("MEM-WR"),
            )
        )
        # Report Idle Status.
        self.sync += self._enable.storage[1].eq(fsm.ongoing("IDLE"))
        # Request Data-Path.
        self.comb += [
            port.source.channel.eq(port.channel),
            port.source.user_id.eq(splitter.source.user_id),
            port.source.first.eq(req_count == 0),
            port.source.last.eq( req_count == (splitter.source.length[length_shift:] - 1)),
            port.source.we.eq(1),
            port.source.adr.eq(splitter.source.address),
            port.source.req_id.eq(endpoint.phy.id),
            port.source.tag.eq(0),
            port.source.len.eq(splitter.source.length[2:]),
            port.source.dat.eq(data_fifo.source.data),
        ]
        # Early termination on last (Optional, can be dynamically disabled).
        self.comb += splitter.terminate.eq(data_fifo.source.last & ~splitter.source.last_disable)

        fsm.act("MEM-WR",
            # Request Control-Path.
            port.source.valid.eq(1),
            # When Request is accepted...
            If(port.source.ready,
                # Increment Request Count.
                NextValue(req_count, req_count + 1),
                # Accept Data (Only when not terminated).
                data_fifo.source.ready.eq(~splitter.terminate),
                # When last...
                If(port.source.last,
                    # Accept Descriptor.
                    splitter.source.ready.eq(1),
                    # Accept Data (Force).
                    data_fifo.source.ready.eq(1),
                    # Return to Idle.
                    NextState("IDLE"),
                )
            )
        )

        # IRQ --------------------------------------------------------------------------------------
        self.comb += If(splitter.source.valid & splitter.source.ready & splitter.source.last,
            self.irq.eq(~splitter.source.irq_disable)
        )

# LitePCIeDMALoopback ------------------------------------------------------------------------------

class LitePCIeDMALoopback(Module, AutoCSR):
    """LitePCIe DMA Loopback

    Optional DMA Reader to DMA Writer loopback.

    For software development or system bring-up/check, being able to do a DMA loopback in the FPGA
    is very useful. This module allows doing a DMA Reader to DMA Writer loopback that can be enabled
    by a CSR. When enabled, user data stream from the DMA Reader is no longer generated, the same
    goes for user data stream to the DMA Writer that is no longer consumed.
    """
    def __init__(self, data_width):
        self.enable      = CSRStorage(description="""DMA Loopback Enable Control.\n
         Write ``1`` to enable DMA internal loopback (DMA Reader to DMA Writer).""")

        self.sink        = stream.Endpoint(dma_layout(data_width))
        self.source      = stream.Endpoint(dma_layout(data_width))

        self.next_source = stream.Endpoint(dma_layout(data_width))
        self.next_sink   = stream.Endpoint(dma_layout(data_width))

        # # #

        self.comb += [
            If(self.enable.storage,
                self.sink.connect(self.source)
            ).Else(
                self.sink.connect(self.next_source),
                self.next_sink.connect(self.source)
            )
        ]

# LitePCIeDMASynchronizer --------------------------------------------------------------------------

class LitePCIeDMASynchronizer(Module, AutoCSR):
    """LitePCIe DMA Synchronizer

    Optional DMA synchronization.

    For some applications (Software Defined Radio, Video, ...), DMA start needs to be precisely
    synchronized to an internal signal of the FPGA (PPS for example for an SDR applications). This
    module allows releasing precisely one or both of the DMA Writer/Reader data streams.
    """
    def __init__(self, data_width):
        self.bypass      = CSRStorage()
        self.enable      = CSRStorage(fields=[
            CSRField("mode", size=2, values=[
                ("``0b00``", "Synchronization disabled."),
                ("``0b01``", "Reader and Writer to PPS Synchronization enabled."),
                ("``0b10``", "PPS Synchronization enabled."),
                ("``0b11``", "Reserved."),
            ]
        )])
        self.ready       = Signal(reset=1)
        self.pps         = Signal()

        self.sink        = stream.Endpoint(dma_layout(data_width))
        self.source      = stream.Endpoint(dma_layout(data_width))

        self.next_source = stream.Endpoint(dma_layout(data_width))
        self.next_sink   = stream.Endpoint(dma_layout(data_width))

        # # #

        self.synced = synced = Signal()

        self.sync += [
            # Synchro Disabled.
            If(self.enable.fields.mode == 0b00,
                synced.eq(0)
            # Synchro Enabled.
            ).Else(
                # On PPS or Bypass and with external ready signal:
                If(self.ready & (self.pps | self.bypass.storage),
                    # TX/RX Synchronization, make sure TX has data.
                    If((self.enable.fields.mode == 0b01) & self.sink.valid,
                        synced.eq(1)
                    ),
                    # Synchronization only on PPS. Reader and Writer are independent
                    If(self.enable.fields.mode == 0b10,
                        synced.eq(1)
                    )
                )
            )
        ]
        self.comb += [
            If(synced,
                self.sink.connect(self.next_source),
                self.next_sink.connect(self.source),
            ).Else(
                # Block sink
                self.next_source.valid.eq(0),
                self.sink.ready.eq(0),

                # Ack next_sink
                self.source.valid.eq(0),
                self.next_sink.ready.eq(1),
            )
        ]

# LitePCIeDMABuffering -----------------------------------------------------------------------------

class LitePCIeDMABuffering(Module, AutoCSR):
    """LitePCIe DMA Buffering

    Optional DMA buffering with dynamically configurable depth.

    For some applications (Software Defined Radio, Video, ...), the user module consuming the data
    from the DMA Reader works at fixed rate and does not handle backpressure. (The same also applies
    to the user module generating the data to the DMA Writer). Since the PCIe bus is shared, gaps
    appears in the streams and our Writes/Reads can't be absorbed/produced at a fixed rate. A minimum
    of buffering is needed to make sure the gaps are smoothed and not propagated to user modules.
    """
    def __init__(self, data_width, with_writer, with_reader, writer_depth, reader_depth, dynamic_depth=True):
        self.sink        = stream.Endpoint(dma_layout(data_width))
        self.source      = stream.Endpoint(dma_layout(data_width))

        self.next_source = stream.Endpoint(dma_layout(data_width))
        self.next_sink   = stream.Endpoint(dma_layout(data_width))

        # Reader FIFO Control/Status.
        if with_reader:
            assert bits_for(reader_depth) < 24
            self.reader_fifo_control = CSRStorage(fields=[
                CSRField("depth", offset=0, size=24, reset=reader_depth,
                    description="DMA Reader FIFO depth (in {}-bit words).".format(data_width)),
                CSRField("scratch",    offset=24, size=4, description="Software Scratchpad."),
                CSRField("level_mode", offset=31, values=[
                    ("``0b0``", "Report Instantaneous level."),
                    ("``0b1``", "Report `Minimal` level since last clear.")
                ])
            ])
            self.reader_fifo_status = CSRStatus(fields=[
                CSRField("level", offset=0, size=24,
                    description="DMA Reader FIFO level (in {}-bit words).".format(data_width))
                ])

        # Writer FIFO Control/Status.
        if with_writer:
            assert bits_for(writer_depth) < 24
            self.writer_fifo_control = CSRStorage(fields=[
                CSRField("depth", offset=0, size=24, reset=writer_depth,
                    description="DMA Writer FIFO depth (in {}-bit words).".format(data_width)),
                CSRField("scratch",    offset=24, size=4, description="Software Scratchpad."),
                CSRField("level_mode", offset=31, values=[
                    ("``0b0``", "Report Instantaneous level."),
                    ("``0b1``", "Report `Maximal` level since last clear.")
                ])
            ])
            self.writer_fifo_status = CSRStatus(fields=[
                CSRField("level", offset=0, size=24,
                    description="DMA Writer FIFO level (in {}-bit words).".format(data_width))
                ])

        # # #

        depth_shift = log2_int(data_width//8)

        # Reader FIFO.
        if with_reader:
            reader_fifo = SyncFIFO(dma_layout(data_width), reader_depth//(data_width//8), buffered=True)
            self.submodules += reader_fifo
            self.comb += [
                # Connect Reader Sink to Reader FIFO when Level < Configured Depth.
                self.sink.connect(reader_fifo.sink, omit={"valid", "ready"}),
                If((reader_fifo.level < self.reader_fifo_control.fields.depth[depth_shift:]) | (not dynamic_depth),
                    self.sink.connect(reader_fifo.sink, keep={"valid", "ready"})
                ),
                # Connect Reader FIFO to Reader Source.
                reader_fifo.source.connect(self.next_source),
            ]

            # Store Min.
            reader_fifo_level_min = Signal.like(reader_fifo.level)
            self.sync += If(reader_fifo.level < reader_fifo_level_min, reader_fifo_level_min.eq(reader_fifo.level))
            # Clear on Status write or when in Instantaneous mode.
            reader_fifo_level_clr = (self.reader_fifo_status.re | (self.reader_fifo_control.fields.level_mode == 0))
            self.sync += If(reader_fifo_level_clr, reader_fifo_level_min.eq(2**len(reader_fifo_level_min)-1))
            # Return Reader FIFO level.
            self.comb += [
                # Instantaneous.
                If(self.reader_fifo_control.fields.level_mode == 0,
                    self.reader_fifo_status.fields.level[depth_shift:].eq(reader_fifo.level)
                # Min.
                ).Else(
                    self.reader_fifo_status.fields.level[depth_shift:].eq(reader_fifo_level_min)
                )
            ]

        # Writer FIFO.
        if with_writer:
            writer_fifo = SyncFIFO(dma_layout(data_width), writer_depth//(data_width//8), buffered=True)
            self.submodules += writer_fifo
            self.comb += [
                # Connect Writer Sink to Writer FIFO when Level < Configured Depth.
                self.next_sink.connect(writer_fifo.sink, omit={"valid", "ready"}),
                If((writer_fifo.level < self.writer_fifo_control.fields.depth[depth_shift:]) | (not dynamic_depth),
                    self.next_sink.connect(writer_fifo.sink, keep={"valid", "ready"})
                ),
                # Connect Writer FIFO to Writer Source.
                writer_fifo.source.connect(self.source),
            ]

            # Store Max.
            writer_fifo_level_max = Signal.like(writer_fifo.level)
            self.sync += If(writer_fifo.level > writer_fifo_level_max, writer_fifo_level_max.eq(writer_fifo.level))
            # Clear on Status write or when in Instantaneous mode.
            writer_fifo_level_clr = (self.writer_fifo_status.re | (self.writer_fifo_control.fields.level_mode == 0))
            self.sync += If(writer_fifo_level_clr, writer_fifo_level_max.eq(0))
            # Return Writer FIFO level.
            self.comb += [
                # Instantaneous.
                If(self.writer_fifo_control.fields.level_mode == 0,
                    self.writer_fifo_status.fields.level[depth_shift:].eq(writer_fifo.level)
                # Min.
                ).Else(
                    self.writer_fifo_status.fields.level[depth_shift:].eq(writer_fifo_level_max)
                )
            ]

# LitePCIeDMAStatus --------------------------------------------------------------------------------

class LitePCIeDMAStatus(Module, AutoCSR):
    """LitePCIe DMA Status

    Optional DMA Status writer to Host Memory.

    LitePCIeDMAStatus writes 16 x 32-bit words to the Host memory. The first 8 words are reserved for
    the internal DMA status and the last 8 words for optional external status. The mapping as follows:

    0:    Sync Word (0x5aa55aa5).
    1:    DMA Writer Loop Status.
    2:    DMA Reader Loop Status.
    3-7:  Reserved
    8-15: External (Optional, from user logic/design).

    The Update to the Host Memory can be triggered from the following events:
    - External (From user logic/design).
    - DMA Writer IRQ.
    - DMA Reader IRQ.
    Allowing a Synchronous or Asynchrounous update with the DMAs.
    """
    def __init__(self, endpoint, writer, reader, address_width=32):
        self.control = CSRStorage(fields=[
            CSRField("enable", offset=0, size=1, description="Status Enable"),
            CSRField("update", offset=4, size=2, description="Status Update Event", values=[
                ("``0b00``", "External."),
                ("``0b01``", "DMA Writer IRQ."),
                ("``0b10``", "DMA Reader IRQ."),
                ("``0b11``", "Software."),
            ]),
        ])
        self.address_lsb = CSRStorage(32, description="Status Base Address (LSB) on Host.")
        self.address_msb = CSRStorage(32, description="Status Base Address (MSB) on Host.")

        self.external_update = Signal()
        self.external_status = Array([Signal(32) for _ in range(8)])

        # # #


        # Create Status Array.
        # --------------------
        status = Array([Signal(32) for _ in range(16)])
        # 0-7:  Internal.
        sync_word = 0x5aa55aa5
        self.comb += [
            status[0].eq(0x5aa55aa5),
            status[1].eq(writer.table.loop_status.status),
            status[2].eq(reader.table.loop_status.status),
        ]

        # 7-15: External.
        for i in range(8):
            self.comb += status[8 + i].eq(self.external_status[i])

        # Update Event.
        # -------------
        update = Signal()
        update_cases = {
            0b00: update.eq(self.external_update),
            0b01: update.eq(writer.irq),
            0b10: update.eq(reader.irq),
            0b11: update.eq(self.control.re),
        }
        self.comb += Case(self.control.fields.update, update_cases)

        # Update Logic.
        # -------------
        port   = endpoint.crossbar.get_master_port(write_only=True)
        dwords = len(port.source.dat)//32
        offset = Signal(4)
        assert len(status)%dwords == 0

        self.submodules.fsm = fsm = FSM(reset_state="IDLE")
        fsm.act("IDLE",
            If(self.control.fields.enable,
                If(update,
                    NextValue(offset, 0),
                    NextState("WORDS-WRITE")
                )
            )
        )
        fsm.act("WORDS-UPDATE",
            NextValue(offset, offset + dwords),
            If(offset == (len(status) - dwords),
                NextState("IDLE")
            ).Else(
                NextState("WORDS-DELAY")
            )
        )
        fsm.act("WORDS-DELAY",
            NextState("WORDS-WRITE")
        )
        self.sync += [
            port.source.channel.eq(port.channel),
            port.source.first.eq(1),
            port.source.last.eq(1),
            port.source.req_id.eq(endpoint.phy.id),
            port.source.tag.eq(0),
            port.source.len.eq(dwords),
            port.source.adr.eq({
                32:              (0x0000_0000 << 32) + self.address_lsb.storage + (offset << 2),
                64: (self.address_msb.storage << 32) + self.address_lsb.storage + (offset << 2),
            }[address_width]),
        ]
        for n in range(dwords):
            self.sync += port.source.dat[32*n:32*(n+1)].eq(status[offset + n])

        fsm.act("WORDS-WRITE",
            port.source.valid.eq(1),
            port.source.we.eq(1),
            If(port.source.ready,
                NextState("WORDS-UPDATE")
            )
        )

# LitePCIeDMA --------------------------------------------------------------------------------------

class LitePCIeDMA(Module, AutoCSR):
    """LitePCIe DMA

    Scatter-Gather bi-directional DMA:
    - Generates a data stream from Host's memory.
    - Stores a data stream to Host's memory.

    Optional buffering, loopback, synchronization and monitoring.
    """
    def __init__(self, phy, endpoint, table_depth=256, address_width=32, with_writer=True, with_reader=True,
        # Loopback.
        with_loopback     = False,
        # Synchronizer.
        with_synchronizer = False,
        # Buffering.
        with_buffering    = False, buffering_depth=256*8, writer_buffering_depth=None, reader_buffering_depth=None,
        # Monitor.
        with_monitor      = False,
        # Status.
        with_status       = False,
        # Splitter Buffer.
        with_writer_splitter_buffer = True,
        with_reader_splitter_buffer = True,
    ):
        # Parameters -------------------------------------------------------------------------------
        self.data_width = data_width = phy.data_width

        # Endoints ---------------------------------------------------------------------------------
        self.sink   = stream.Endpoint(dma_layout(data_width))
        self.source = stream.Endpoint(dma_layout(data_width))

        # Writer/Reader ----------------------------------------------------------------------------
        if with_writer:
            writer = LitePCIeDMAWriter(
                endpoint             = endpoint,
                port                 = endpoint.crossbar.get_master_port(write_only=True),
                table_depth          = table_depth,
                address_width        = address_width,
                with_splitter_buffer = with_writer_splitter_buffer,
            )
            self.submodules.writer = writer
            self.comb += self.sink.connect(self.writer.sink)

        if with_reader:
            reader = LitePCIeDMAReader(
                endpoint             = endpoint,
                port                 = endpoint.crossbar.get_master_port(read_only=True),
                table_depth          = table_depth,
                address_width        = address_width,
                with_splitter_buffer = with_reader_splitter_buffer,
            )
            self.submodules.reader = reader
            self.comb += reader.source.connect(self.source)

        # Loopback ---------------------------------------------------------------------------------
        if with_loopback:
            if not (with_writer and with_reader):
                raise ValueError("Loopback capability requires DMAWriter and DMAReader to be enabled.")
            self.submodules.loopback = LitePCIeDMALoopback(data_width)
            self.add_plugin_module(self.loopback)

        # Synchronizer -----------------------------------------------------------------------------
        if with_synchronizer:
            if not (with_writer and with_reader):
                raise ValueError("Synchronizer capability requires DMAWriter and DMAReader to be enabled.")
            self.submodules.synchronizer = LitePCIeDMASynchronizer(data_width)
            self.add_plugin_module(self.synchronizer)

        # Buffering --------------------------------------------------------------------------------
        if with_buffering:
            writer_depth = writer_buffering_depth if writer_buffering_depth is not None else buffering_depth
            reader_depth = reader_buffering_depth if reader_buffering_depth is not None else buffering_depth
            self.submodules.buffering = LitePCIeDMABuffering(
                data_width   = data_width,
                with_reader  = with_reader,
                with_writer  = with_writer,
                reader_depth = reader_depth,
                writer_depth = writer_depth,
            )
            self.add_plugin_module(self.buffering)

        # Monitor ----------------------------------------------------------------------------------
        if with_monitor:
            if with_writer:
                self.submodules.writer_monitor = stream.Monitor(self.sink,   count_width=16, with_overflows  = True)
            if with_reader:
                self.submodules.reader_monitor = stream.Monitor(self.source, count_width=16, with_underflows = True)

        # Status -----------------------------------------------------------------------------------
        if with_status:
            if not (with_writer and with_reader):
                raise ValueError("Status capability requires DMAWriter and DMAReader to be enabled.")
            self.submodules.status = LitePCIeDMAStatus(
                endpoint      = endpoint,
                writer        = writer,
                reader        = reader,
                address_width = address_width,
            )

    def add_plugin_module(self, m):
        self.comb += [
            self.source.connect(m.sink),
            m.source.connect(self.sink)
        ]
        self.sink, self.source = m.next_sink, m.next_source



import os
from distutils.dir_util import copy_tree

from litex.build import tools

from litex.soc.integration.export import get_csr_header, get_soc_header, get_mem_header


def copy_litepcie_software(dst):
    src = os.path.abspath(os.path.dirname(__file__))
    copy_tree(src, dst)

def generate_litepcie_software_headers(soc, dst):
    csr_header = get_csr_header(soc.csr_regions, soc.constants, with_access_functions=False)
    tools.write_to_file(os.path.join(dst, "csr.h"), csr_header)
    soc_header = get_soc_header(soc.constants, with_access_functions=False)
    tools.write_to_file(os.path.join(dst, "soc.h"), soc_header)
    mem_header = get_mem_header(soc.mem_regions)
    tools.write_to_file(os.path.join(dst, "mem.h"), mem_header)

def generate_litepcie_software(soc, dst):
    copy_litepcie_software(dst)
    generate_litepcie_software_headers(soc, os.path.join(dst, "kernel"))



#
# This file is part of LitePCIe.
#
# Copyright (c) 2015-2022 Florent Kermarrec <florent@enjoy-digital.fr>
# SPDX-License-Identifier: BSD-2-Clause

from migen import *

from litex.soc.interconnect.csr import *

from litepcie.tlp.depacketizer import LitePCIeTLPDepacketizer
from litepcie.tlp.packetizer import LitePCIeTLPPacketizer
from litepcie.core.crossbar import LitePCIeCrossbar

# LitePCIe Endpoint --------------------------------------------------------------------------------

class LitePCIeEndpoint(Module):
    def __init__(self, phy, max_pending_requests=4, address_width=32, endianness="big", cmp_bufs_buffered=True):
        self.phy                  = phy
        self.max_pending_requests = max_pending_requests

        # # #

        # TLP Packetizer / Depacketizer ------------------------------------------------------------
        if hasattr(phy, "sink") and hasattr(phy, "source"):
            # Shared Request/Completion channels
            depacketizer = LitePCIeTLPDepacketizer(phy.data_width, endianness, phy.bar0_mask)
            packetizer   = LitePCIeTLPPacketizer(phy.data_width, endianness, address_width)
            self.submodules.depacketizer = depacketizer
            self.submodules.packetizer   = packetizer
            self.comb += [
                phy.source.connect(depacketizer.sink),
                packetizer.source.connect(phy.sink)
            ]
            req_source = depacketizer.req_source
            cmp_sink   = packetizer.cmp_sink
            req_sink   = packetizer.req_sink
            cmp_source = depacketizer.cmp_source
        else:
            # Separate Request/Completion channels
            cmp_depacketizer = LitePCIeTLPDepacketizer(phy.data_width, endianness, phy.bar0_mask)
            req_depacketizer = LitePCIeTLPDepacketizer(phy.data_width, endianness, phy.bar0_mask)
            cmp_packetizer   = LitePCIeTLPPacketizer(phy.data_width, endianness, address_width)
            req_packetizer   = LitePCIeTLPPacketizer(phy.data_width, endianness, address_width)
            self.submodules.cmp_depacketizer = cmp_depacketizer
            self.submodules.req_depacketizer = req_depacketizer
            self.submodules.cmp_packetizer   = cmp_packetizer
            self.submodules.req_packetizer   = req_packetizer
            self.comb += [
                phy.cmp_source.connect(cmp_depacketizer.sink),
                phy.req_source.connect(req_depacketizer.sink),
                cmp_packetizer.source.connect(phy.cmp_sink),
                req_packetizer.source.connect(phy.req_sink),
            ]
            req_source = req_depacketizer.req_source
            cmp_sink   = cmp_packetizer.cmp_sink
            req_sink   = req_packetizer.req_sink
            cmp_source = cmp_depacketizer.cmp_source

        # Crossbar ---------------------------------------------------------------------------------
        crossbar = LitePCIeCrossbar(
            data_width           = phy.data_width,
            address_width        = address_width,
            max_pending_requests = max_pending_requests,
            cmp_bufs_buffered    = cmp_bufs_buffered
        )
        self.submodules.crossbar = crossbar

        # Slave: HOST initiates the transactions ---------------------------------------------------
        self.comb += [
            req_source.connect(crossbar.phy_slave.sink),
            crossbar.phy_slave.source.connect(cmp_sink),
        ]

        # Master: FPGA initiates the transactions --------------------------------------------------
        self.comb += [
            crossbar.phy_master.source.connect(req_sink),
            cmp_source.connect(crossbar.phy_master.sink),
        ]
